// Copyright Take Vos 2022.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// https://www.boost.org/LICENSE_1_0.txt)

#include "language_tag.hpp"
#include "../check.hpp"
#include <algorithm>
#include <numeric>

namespace hi::inline v1 {

struct language_tag_expansion {
    std::string_view from;
    std::string_view to;

    constexpr language_tag_expansion(std::string_view from, std::string_view to) noexcept : from(from), to(to) {}
};

// clang-format off
/** Table of likely expansions.
 * 
 * Most data from this table comes from unicode locale database:
 *    https://github.com/unicode-org/cldr/blob/release-40/common/supplemental/likelySubtags.xml
 * 
 * Also included is a modified grandfathered table.
 */
constexpr auto language_tag_expansions = std::array{
#ifndef __INTELLISENSE__
    language_tag_expansion{"aa", "aa-Latn-ET"},
    language_tag_expansion{"aai", "aai-Latn-ZZ"},
    language_tag_expansion{"aak", "aak-Latn-ZZ"},
    language_tag_expansion{"aau", "aau-Latn-ZZ"},
    language_tag_expansion{"ab", "ab-Cyrl-GE"},
    language_tag_expansion{"abi", "abi-Latn-ZZ"},
    language_tag_expansion{"abq", "abq-Cyrl-ZZ"},
    language_tag_expansion{"abr", "abr-Latn-GH"},
    language_tag_expansion{"abt", "abt-Latn-ZZ"},
    language_tag_expansion{"aby", "aby-Latn-ZZ"},
    language_tag_expansion{"acd", "acd-Latn-ZZ"},
    language_tag_expansion{"ace", "ace-Latn-ID"},
    language_tag_expansion{"ach", "ach-Latn-UG"},
    language_tag_expansion{"ada", "ada-Latn-GH"},
    language_tag_expansion{"ade", "ade-Latn-ZZ"},
    language_tag_expansion{"adj", "adj-Latn-ZZ"},
    language_tag_expansion{"adp", "adp-Tibt-BT"},
    language_tag_expansion{"ady", "ady-Cyrl-RU"},
    language_tag_expansion{"adz", "adz-Latn-ZZ"},
    language_tag_expansion{"ae", "ae-Avst-IR"},
    language_tag_expansion{"aeb", "aeb-Arab-TN"},
    language_tag_expansion{"aey", "aey-Latn-ZZ"},
    language_tag_expansion{"af", "af-Latn-ZA"},
    language_tag_expansion{"agc", "agc-Latn-ZZ"},
    language_tag_expansion{"agd", "agd-Latn-ZZ"},
    language_tag_expansion{"agg", "agg-Latn-ZZ"},
    language_tag_expansion{"agm", "agm-Latn-ZZ"},
    language_tag_expansion{"ago", "ago-Latn-ZZ"},
    language_tag_expansion{"agq", "agq-Latn-CM"},
    language_tag_expansion{"aha", "aha-Latn-ZZ"},
    language_tag_expansion{"ahl", "ahl-Latn-ZZ"},
    language_tag_expansion{"aho", "aho-Ahom-IN"},
    language_tag_expansion{"ajg", "ajg-Latn-ZZ"},
    language_tag_expansion{"ak", "ak-Latn-GH"},
    language_tag_expansion{"akk", "akk-Xsux-IQ"},
    language_tag_expansion{"ala", "ala-Latn-ZZ"},
    language_tag_expansion{"ali", "ali-Latn-ZZ"},
    language_tag_expansion{"aln", "aln-Latn-XK"},
    language_tag_expansion{"alt", "alt-Cyrl-RU"},
    language_tag_expansion{"am", "am-Ethi-ET"},
    language_tag_expansion{"amm", "amm-Latn-ZZ"},
    language_tag_expansion{"amn", "amn-Latn-ZZ"},
    language_tag_expansion{"amo", "amo-Latn-NG"},
    language_tag_expansion{"amp", "amp-Latn-ZZ"},
    language_tag_expansion{"an", "an-Latn-ES"},
    language_tag_expansion{"anc", "anc-Latn-ZZ"},
    language_tag_expansion{"ank", "ank-Latn-ZZ"},
    language_tag_expansion{"ann", "ann-Latn-ZZ"},
    language_tag_expansion{"any", "any-Latn-ZZ"},
    language_tag_expansion{"aoj", "aoj-Latn-ZZ"},
    language_tag_expansion{"aom", "aom-Latn-ZZ"},
    language_tag_expansion{"aoz", "aoz-Latn-ID"},
    language_tag_expansion{"apc", "apc-Arab-ZZ"},
    language_tag_expansion{"apd", "apd-Arab-TG"},
    language_tag_expansion{"ape", "ape-Latn-ZZ"},
    language_tag_expansion{"apr", "apr-Latn-ZZ"},
    language_tag_expansion{"aps", "aps-Latn-ZZ"},
    language_tag_expansion{"apz", "apz-Latn-ZZ"},
    language_tag_expansion{"ar", "ar-Arab-EG"},
    language_tag_expansion{"arc", "arc-Armi-IR"},
    language_tag_expansion{"arc-nbat", "arc-Nbat-JO"},
    language_tag_expansion{"arc-palm", "arc-Palm-SY"},
    language_tag_expansion{"arh", "arh-Latn-ZZ"},
    language_tag_expansion{"arn", "arn-Latn-CL"},
    language_tag_expansion{"aro", "aro-Latn-BO"},
    language_tag_expansion{"arq", "arq-Arab-DZ"},
    language_tag_expansion{"ars", "ars-Arab-SA"},
    language_tag_expansion{"ary", "ary-Arab-MA"},
    language_tag_expansion{"arz", "arz-Arab-EG"},
    language_tag_expansion{"as", "as-Beng-IN"},
    language_tag_expansion{"asa", "asa-Latn-TZ"},
    language_tag_expansion{"ase", "ase-Sgnw-US"},
    language_tag_expansion{"asg", "asg-Latn-ZZ"},
    language_tag_expansion{"aso", "aso-Latn-ZZ"},
    language_tag_expansion{"ast", "ast-Latn-ES"},
    language_tag_expansion{"ata", "ata-Latn-ZZ"},
    language_tag_expansion{"atg", "atg-Latn-ZZ"},
    language_tag_expansion{"atj", "atj-Latn-CA"},
    language_tag_expansion{"auy", "auy-Latn-ZZ"},
    language_tag_expansion{"av", "av-Cyrl-RU"},
    language_tag_expansion{"avl", "avl-Arab-ZZ"},
    language_tag_expansion{"avn", "avn-Latn-ZZ"},
    language_tag_expansion{"avt", "avt-Latn-ZZ"},
    language_tag_expansion{"avu", "avu-Latn-ZZ"},
    language_tag_expansion{"awa", "awa-Deva-IN"},
    language_tag_expansion{"awb", "awb-Latn-ZZ"},
    language_tag_expansion{"awo", "awo-Latn-ZZ"},
    language_tag_expansion{"awx", "awx-Latn-ZZ"},
    language_tag_expansion{"ay", "ay-Latn-BO"},
    language_tag_expansion{"ayb", "ayb-Latn-ZZ"},
    language_tag_expansion{"az", "az-Latn-AZ"},
    language_tag_expansion{"az-arab", "az-Arab-IR"},
    language_tag_expansion{"az-iq", "az-Arab-IQ"},
    language_tag_expansion{"az-ir", "az-Arab-IR"},
    language_tag_expansion{"az-ru", "az-Cyrl-RU"},
    language_tag_expansion{"ba", "ba-Cyrl-RU"},
    language_tag_expansion{"bal", "bal-Arab-PK"},
    language_tag_expansion{"ban", "ban-Latn-ID"},
    language_tag_expansion{"bap", "bap-Deva-NP"},
    language_tag_expansion{"bar", "bar-Latn-AT"},
    language_tag_expansion{"bas", "bas-Latn-CM"},
    language_tag_expansion{"bav", "bav-Latn-ZZ"},
    language_tag_expansion{"bax", "bax-Bamu-CM"},
    language_tag_expansion{"bba", "bba-Latn-ZZ"},
    language_tag_expansion{"bbb", "bbb-Latn-ZZ"},
    language_tag_expansion{"bbc", "bbc-Latn-ID"},
    language_tag_expansion{"bbd", "bbd-Latn-ZZ"},
    language_tag_expansion{"bbj", "bbj-Latn-CM"},
    language_tag_expansion{"bbp", "bbp-Latn-ZZ"},
    language_tag_expansion{"bbr", "bbr-Latn-ZZ"},
    language_tag_expansion{"bcf", "bcf-Latn-ZZ"},
    language_tag_expansion{"bch", "bch-Latn-ZZ"},
    language_tag_expansion{"bci", "bci-Latn-CI"},
    language_tag_expansion{"bcm", "bcm-Latn-ZZ"},
    language_tag_expansion{"bcn", "bcn-Latn-ZZ"},
    language_tag_expansion{"bco", "bco-Latn-ZZ"},
    language_tag_expansion{"bcq", "bcq-Ethi-ZZ"},
    language_tag_expansion{"bcu", "bcu-Latn-ZZ"},
    language_tag_expansion{"bdd", "bdd-Latn-ZZ"},
    language_tag_expansion{"be", "be-Cyrl-BY"},
    language_tag_expansion{"bef", "bef-Latn-ZZ"},
    language_tag_expansion{"beh", "beh-Latn-ZZ"},
    language_tag_expansion{"bej", "bej-Arab-SD"},
    language_tag_expansion{"bem", "bem-Latn-ZM"},
    language_tag_expansion{"bet", "bet-Latn-ZZ"},
    language_tag_expansion{"bew", "bew-Latn-ID"},
    language_tag_expansion{"bex", "bex-Latn-ZZ"},
    language_tag_expansion{"bez", "bez-Latn-TZ"},
    language_tag_expansion{"bfd", "bfd-Latn-CM"},
    language_tag_expansion{"bfq", "bfq-Taml-IN"},
    language_tag_expansion{"bft", "bft-Arab-PK"},
    language_tag_expansion{"bfy", "bfy-Deva-IN"},
    language_tag_expansion{"bg", "bg-Cyrl-BG"},
    language_tag_expansion{"bgc", "bgc-Deva-IN"},
    language_tag_expansion{"bgn", "bgn-Arab-PK"},
    language_tag_expansion{"bgx", "bgx-Grek-TR"},
    language_tag_expansion{"bhb", "bhb-Deva-IN"},
    language_tag_expansion{"bhg", "bhg-Latn-ZZ"},
    language_tag_expansion{"bhi", "bhi-Deva-IN"},
    language_tag_expansion{"bhl", "bhl-Latn-ZZ"},
    language_tag_expansion{"bho", "bho-Deva-IN"},
    language_tag_expansion{"bhy", "bhy-Latn-ZZ"},
    language_tag_expansion{"bi", "bi-Latn-VU"},
    language_tag_expansion{"bib", "bib-Latn-ZZ"},
    language_tag_expansion{"big", "big-Latn-ZZ"},
    language_tag_expansion{"bik", "bik-Latn-PH"},
    language_tag_expansion{"bim", "bim-Latn-ZZ"},
    language_tag_expansion{"bin", "bin-Latn-NG"},
    language_tag_expansion{"bio", "bio-Latn-ZZ"},
    language_tag_expansion{"biq", "biq-Latn-ZZ"},
    language_tag_expansion{"bjh", "bjh-Latn-ZZ"},
    language_tag_expansion{"bji", "bji-Ethi-ZZ"},
    language_tag_expansion{"bjj", "bjj-Deva-IN"},
    language_tag_expansion{"bjn", "bjn-Latn-ID"},
    language_tag_expansion{"bjo", "bjo-Latn-ZZ"},
    language_tag_expansion{"bjr", "bjr-Latn-ZZ"},
    language_tag_expansion{"bjt", "bjt-Latn-SN"},
    language_tag_expansion{"bjz", "bjz-Latn-ZZ"},
    language_tag_expansion{"bkc", "bkc-Latn-ZZ"},
    language_tag_expansion{"bkm", "bkm-Latn-CM"},
    language_tag_expansion{"bkq", "bkq-Latn-ZZ"},
    language_tag_expansion{"bku", "bku-Latn-PH"},
    language_tag_expansion{"bkv", "bkv-Latn-ZZ"},
    language_tag_expansion{"blg", "blg-Latn-MY"},
    language_tag_expansion{"blt", "blt-Tavt-VN"},
    language_tag_expansion{"bm", "bm-Latn-ML"},
    language_tag_expansion{"bmh", "bmh-Latn-ZZ"},
    language_tag_expansion{"bmk", "bmk-Latn-ZZ"},
    language_tag_expansion{"bmq", "bmq-Latn-ML"},
    language_tag_expansion{"bmu", "bmu-Latn-ZZ"},
    language_tag_expansion{"bn", "bn-Beng-BD"},
    language_tag_expansion{"bng", "bng-Latn-ZZ"},
    language_tag_expansion{"bnm", "bnm-Latn-ZZ"},
    language_tag_expansion{"bnp", "bnp-Latn-ZZ"},
    language_tag_expansion{"bo", "bo-Tibt-CN"},
    language_tag_expansion{"boj", "boj-Latn-ZZ"},
    language_tag_expansion{"bom", "bom-Latn-ZZ"},
    language_tag_expansion{"bon", "bon-Latn-ZZ"},
    language_tag_expansion{"bpy", "bpy-Beng-IN"},
    language_tag_expansion{"bqc", "bqc-Latn-ZZ"},
    language_tag_expansion{"bqi", "bqi-Arab-IR"},
    language_tag_expansion{"bqp", "bqp-Latn-ZZ"},
    language_tag_expansion{"bqv", "bqv-Latn-CI"},
    language_tag_expansion{"br", "br-Latn-FR"},
    language_tag_expansion{"bra", "bra-Deva-IN"},
    language_tag_expansion{"brh", "brh-Arab-PK"},
    language_tag_expansion{"brx", "brx-Deva-IN"},
    language_tag_expansion{"brz", "brz-Latn-ZZ"},
    language_tag_expansion{"bs", "bs-Latn-BA"},
    language_tag_expansion{"bsj", "bsj-Latn-ZZ"},
    language_tag_expansion{"bsq", "bsq-Bass-LR"},
    language_tag_expansion{"bss", "bss-Latn-CM"},
    language_tag_expansion{"bst", "bst-Ethi-ZZ"},
    language_tag_expansion{"bto", "bto-Latn-PH"},
    language_tag_expansion{"btt", "btt-Latn-ZZ"},
    language_tag_expansion{"btv", "btv-Deva-PK"},
    language_tag_expansion{"bua", "bua-Cyrl-RU"},
    language_tag_expansion{"buc", "buc-Latn-YT"},
    language_tag_expansion{"bud", "bud-Latn-ZZ"},
    language_tag_expansion{"bug", "bug-Latn-ID"},
    language_tag_expansion{"buk", "buk-Latn-ZZ"},
    language_tag_expansion{"bum", "bum-Latn-CM"},
    language_tag_expansion{"buo", "buo-Latn-ZZ"},
    language_tag_expansion{"bus", "bus-Latn-ZZ"},
    language_tag_expansion{"buu", "buu-Latn-ZZ"},
    language_tag_expansion{"bvb", "bvb-Latn-GQ"},
    language_tag_expansion{"bwd", "bwd-Latn-ZZ"},
    language_tag_expansion{"bwr", "bwr-Latn-ZZ"},
    language_tag_expansion{"bxh", "bxh-Latn-ZZ"},
    language_tag_expansion{"bye", "bye-Latn-ZZ"},
    language_tag_expansion{"byn", "byn-Ethi-ER"},
    language_tag_expansion{"byr", "byr-Latn-ZZ"},
    language_tag_expansion{"bys", "bys-Latn-ZZ"},
    language_tag_expansion{"byv", "byv-Latn-CM"},
    language_tag_expansion{"byx", "byx-Latn-ZZ"},
    language_tag_expansion{"bza", "bza-Latn-ZZ"},
    language_tag_expansion{"bze", "bze-Latn-ML"},
    language_tag_expansion{"bzf", "bzf-Latn-ZZ"},
    language_tag_expansion{"bzh", "bzh-Latn-ZZ"},
    language_tag_expansion{"bzw", "bzw-Latn-ZZ"},
    language_tag_expansion{"ca", "ca-Latn-ES"},
    language_tag_expansion{"cad", "cad-Latn-US"},
    language_tag_expansion{"can", "can-Latn-ZZ"},
    language_tag_expansion{"cbj", "cbj-Latn-ZZ"},
    language_tag_expansion{"cch", "cch-Latn-NG"},
    language_tag_expansion{"ccp", "ccp-Cakm-BD"},
    language_tag_expansion{"ce", "ce-Cyrl-RU"},
    language_tag_expansion{"ceb", "ceb-Latn-PH"},
    language_tag_expansion{"cfa", "cfa-Latn-ZZ"},
    language_tag_expansion{"cgg", "cgg-Latn-UG"},
    language_tag_expansion{"ch", "ch-Latn-GU"},
    language_tag_expansion{"chk", "chk-Latn-FM"},
    language_tag_expansion{"chm", "chm-Cyrl-RU"},
    language_tag_expansion{"cho", "cho-Latn-US"},
    language_tag_expansion{"chp", "chp-Latn-CA"},
    language_tag_expansion{"chr", "chr-Cher-US"},
    language_tag_expansion{"cic", "cic-Latn-US"},
    language_tag_expansion{"cja", "cja-Arab-KH"},
    language_tag_expansion{"cjm", "cjm-Cham-VN"},
    language_tag_expansion{"cjv", "cjv-Latn-ZZ"},
    language_tag_expansion{"ckb", "ckb-Arab-IQ"},
    language_tag_expansion{"ckl", "ckl-Latn-ZZ"},
    language_tag_expansion{"cko", "cko-Latn-ZZ"},
    language_tag_expansion{"cky", "cky-Latn-ZZ"},
    language_tag_expansion{"cla", "cla-Latn-ZZ"},
    language_tag_expansion{"cme", "cme-Latn-ZZ"},
    language_tag_expansion{"cmg", "cmg-Soyo-MN"},
    language_tag_expansion{"co", "co-Latn-FR"},
    language_tag_expansion{"cop", "cop-Copt-EG"},
    language_tag_expansion{"cps", "cps-Latn-PH"},
    language_tag_expansion{"cr", "cr-Cans-CA"},
    language_tag_expansion{"crh", "crh-Cyrl-UA"},
    language_tag_expansion{"crj", "crj-Cans-CA"},
    language_tag_expansion{"crk", "crk-Cans-CA"},
    language_tag_expansion{"crl", "crl-Cans-CA"},
    language_tag_expansion{"crm", "crm-Cans-CA"},
    language_tag_expansion{"crs", "crs-Latn-SC"},
    language_tag_expansion{"cs", "cs-Latn-CZ"},
    language_tag_expansion{"csb", "csb-Latn-PL"},
    language_tag_expansion{"csw", "csw-Cans-CA"},
    language_tag_expansion{"ctd", "ctd-Pauc-MM"},
    language_tag_expansion{"cu", "cu-Cyrl-RU"},
    language_tag_expansion{"cu-glag", "cu-Glag-BG"},
    language_tag_expansion{"cv", "cv-Cyrl-RU"},
    language_tag_expansion{"cy", "cy-Latn-GB"},
    language_tag_expansion{"da", "da-Latn-DK"},
    language_tag_expansion{"dad", "dad-Latn-ZZ"},
    language_tag_expansion{"daf", "daf-Latn-CI"},
    language_tag_expansion{"dag", "dag-Latn-ZZ"},
    language_tag_expansion{"dah", "dah-Latn-ZZ"},
    language_tag_expansion{"dak", "dak-Latn-US"},
    language_tag_expansion{"dar", "dar-Cyrl-RU"},
    language_tag_expansion{"dav", "dav-Latn-KE"},
    language_tag_expansion{"dbd", "dbd-Latn-ZZ"},
    language_tag_expansion{"dbq", "dbq-Latn-ZZ"},
    language_tag_expansion{"dcc", "dcc-Arab-IN"},
    language_tag_expansion{"ddn", "ddn-Latn-ZZ"},
    language_tag_expansion{"de", "de-Latn-DE"},
    language_tag_expansion{"ded", "ded-Latn-ZZ"},
    language_tag_expansion{"den", "den-Latn-CA"},
    language_tag_expansion{"dga", "dga-Latn-ZZ"},
    language_tag_expansion{"dgh", "dgh-Latn-ZZ"},
    language_tag_expansion{"dgi", "dgi-Latn-ZZ"},
    language_tag_expansion{"dgl", "dgl-Arab-ZZ"},
    language_tag_expansion{"dgr", "dgr-Latn-CA"},
    language_tag_expansion{"dgz", "dgz-Latn-ZZ"},
    language_tag_expansion{"dia", "dia-Latn-ZZ"},
    language_tag_expansion{"dje", "dje-Latn-NE"},
    language_tag_expansion{"dmf", "dmf-Medf-NG"},
    language_tag_expansion{"dnj", "dnj-Latn-CI"},
    language_tag_expansion{"dob", "dob-Latn-ZZ"},
    language_tag_expansion{"doi", "doi-Deva-IN"},
    language_tag_expansion{"dop", "dop-Latn-ZZ"},
    language_tag_expansion{"dow", "dow-Latn-ZZ"},
    language_tag_expansion{"drh", "drh-Mong-CN"},
    language_tag_expansion{"dri", "dri-Latn-ZZ"},
    language_tag_expansion{"drs", "drs-Ethi-ZZ"},
    language_tag_expansion{"dsb", "dsb-Latn-DE"},
    language_tag_expansion{"dtm", "dtm-Latn-ML"},
    language_tag_expansion{"dtp", "dtp-Latn-MY"},
    language_tag_expansion{"dts", "dts-Latn-ZZ"},
    language_tag_expansion{"dty", "dty-Deva-NP"},
    language_tag_expansion{"dua", "dua-Latn-CM"},
    language_tag_expansion{"duc", "duc-Latn-ZZ"},
    language_tag_expansion{"dud", "dud-Latn-ZZ"},
    language_tag_expansion{"dug", "dug-Latn-ZZ"},
    language_tag_expansion{"dv", "dv-Thaa-MV"},
    language_tag_expansion{"dva", "dva-Latn-ZZ"},
    language_tag_expansion{"dww", "dww-Latn-ZZ"},
    language_tag_expansion{"dyo", "dyo-Latn-SN"},
    language_tag_expansion{"dyu", "dyu-Latn-BF"},
    language_tag_expansion{"dz", "dz-Tibt-BT"},
    language_tag_expansion{"dzg", "dzg-Latn-ZZ"},
    language_tag_expansion{"ebu", "ebu-Latn-KE"},
    language_tag_expansion{"ee", "ee-Latn-GH"},
    language_tag_expansion{"efi", "efi-Latn-NG"},
    language_tag_expansion{"egl", "egl-Latn-IT"},
    language_tag_expansion{"egy", "egy-Egyp-EG"},
    language_tag_expansion{"eka", "eka-Latn-ZZ"},
    language_tag_expansion{"eky", "eky-Kali-MM"},
    language_tag_expansion{"el", "el-Grek-GR"},
    language_tag_expansion{"ema", "ema-Latn-ZZ"},
    language_tag_expansion{"emi", "emi-Latn-ZZ"},
    language_tag_expansion{"en", "en-Latn-US"},
    language_tag_expansion{"en-gb-oed", "en-GB-oxendict"},
    language_tag_expansion{"enn", "enn-Latn-ZZ"},
    language_tag_expansion{"enq", "enq-Latn-ZZ"},
    language_tag_expansion{"en-shaw", "en-Shaw-GB"},
    language_tag_expansion{"eo", "eo-Latn-001"},
    language_tag_expansion{"eri", "eri-Latn-ZZ"},
    language_tag_expansion{"es", "es-Latn-ES"},
    language_tag_expansion{"esg", "esg-Gonm-IN"},
    language_tag_expansion{"esu", "esu-Latn-US"},
    language_tag_expansion{"et", "et-Latn-EE"},
    language_tag_expansion{"etr", "etr-Latn-ZZ"},
    language_tag_expansion{"ett", "ett-Ital-IT"},
    language_tag_expansion{"etu", "etu-Latn-ZZ"},
    language_tag_expansion{"etx", "etx-Latn-ZZ"},
    language_tag_expansion{"eu", "eu-Latn-ES"},
    language_tag_expansion{"ewo", "ewo-Latn-CM"},
    language_tag_expansion{"ext", "ext-Latn-ES"},
    language_tag_expansion{"eza", "eza-Latn-ZZ"},
    language_tag_expansion{"fa", "fa-Arab-IR"},
    language_tag_expansion{"faa", "faa-Latn-ZZ"},
    language_tag_expansion{"fab", "fab-Latn-ZZ"},
    language_tag_expansion{"fag", "fag-Latn-ZZ"},
    language_tag_expansion{"fai", "fai-Latn-ZZ"},
    language_tag_expansion{"fan", "fan-Latn-GQ"},
    language_tag_expansion{"ff", "ff-Latn-SN"},
    language_tag_expansion{"ff-adlm", "ff-Adlm-GN"},
    language_tag_expansion{"ffi", "ffi-Latn-ZZ"},
    language_tag_expansion{"ffm", "ffm-Latn-ML"},
    language_tag_expansion{"fi", "fi-Latn-FI"},
    language_tag_expansion{"fia", "fia-Arab-SD"},
    language_tag_expansion{"fil", "fil-Latn-PH"},
    language_tag_expansion{"fit", "fit-Latn-SE"},
    language_tag_expansion{"fj", "fj-Latn-FJ"},
    language_tag_expansion{"flr", "flr-Latn-ZZ"},
    language_tag_expansion{"fmp", "fmp-Latn-ZZ"},
    language_tag_expansion{"fo", "fo-Latn-FO"},
    language_tag_expansion{"fod", "fod-Latn-ZZ"},
    language_tag_expansion{"fon", "fon-Latn-BJ"},
    language_tag_expansion{"for", "for-Latn-ZZ"},
    language_tag_expansion{"fpe", "fpe-Latn-ZZ"},
    language_tag_expansion{"fqs", "fqs-Latn-ZZ"},
    language_tag_expansion{"fr", "fr-Latn-FR"},
    language_tag_expansion{"frc", "frc-Latn-US"},
    language_tag_expansion{"frp", "frp-Latn-FR"},
    language_tag_expansion{"frr", "frr-Latn-DE"},
    language_tag_expansion{"frs", "frs-Latn-DE"},
    language_tag_expansion{"fub", "fub-Arab-CM"},
    language_tag_expansion{"fud", "fud-Latn-WF"},
    language_tag_expansion{"fue", "fue-Latn-ZZ"},
    language_tag_expansion{"fuf", "fuf-Latn-GN"},
    language_tag_expansion{"fuh", "fuh-Latn-ZZ"},
    language_tag_expansion{"fuq", "fuq-Latn-NE"},
    language_tag_expansion{"fur", "fur-Latn-IT"},
    language_tag_expansion{"fuv", "fuv-Latn-NG"},
    language_tag_expansion{"fuy", "fuy-Latn-ZZ"},
    language_tag_expansion{"fvr", "fvr-Latn-SD"},
    language_tag_expansion{"fy", "fy-Latn-NL"},
    language_tag_expansion{"ga", "ga-Latn-IE"},
    language_tag_expansion{"gaa", "gaa-Latn-GH"},
    language_tag_expansion{"gaf", "gaf-Latn-ZZ"},
    language_tag_expansion{"gag", "gag-Latn-MD"},
    language_tag_expansion{"gah", "gah-Latn-ZZ"},
    language_tag_expansion{"gaj", "gaj-Latn-ZZ"},
    language_tag_expansion{"gam", "gam-Latn-ZZ"},
    language_tag_expansion{"gan", "gan-Hans-CN"},
    language_tag_expansion{"gaw", "gaw-Latn-ZZ"},
    language_tag_expansion{"gay", "gay-Latn-ID"},
    language_tag_expansion{"gba", "gba-Latn-ZZ"},
    language_tag_expansion{"gbf", "gbf-Latn-ZZ"},
    language_tag_expansion{"gbm", "gbm-Deva-IN"},
    language_tag_expansion{"gby", "gby-Latn-ZZ"},
    language_tag_expansion{"gbz", "gbz-Arab-IR"},
    language_tag_expansion{"gcr", "gcr-Latn-GF"},
    language_tag_expansion{"gd", "gd-Latn-GB"},
    language_tag_expansion{"gde", "gde-Latn-ZZ"},
    language_tag_expansion{"gdn", "gdn-Latn-ZZ"},
    language_tag_expansion{"gdr", "gdr-Latn-ZZ"},
    language_tag_expansion{"geb", "geb-Latn-ZZ"},
    language_tag_expansion{"gej", "gej-Latn-ZZ"},
    language_tag_expansion{"gel", "gel-Latn-ZZ"},
    language_tag_expansion{"gez", "gez-Ethi-ET"},
    language_tag_expansion{"gfk", "gfk-Latn-ZZ"},
    language_tag_expansion{"ggn", "ggn-Deva-NP"},
    language_tag_expansion{"ghs", "ghs-Latn-ZZ"},
    language_tag_expansion{"gil", "gil-Latn-KI"},
    language_tag_expansion{"gim", "gim-Latn-ZZ"},
    language_tag_expansion{"gjk", "gjk-Arab-PK"},
    language_tag_expansion{"gjn", "gjn-Latn-ZZ"},
    language_tag_expansion{"gju", "gju-Arab-PK"},
    language_tag_expansion{"gkn", "gkn-Latn-ZZ"},
    language_tag_expansion{"gkp", "gkp-Latn-ZZ"},
    language_tag_expansion{"gl", "gl-Latn-ES"},
    language_tag_expansion{"glk", "glk-Arab-IR"},
    language_tag_expansion{"gmm", "gmm-Latn-ZZ"},
    language_tag_expansion{"gmv", "gmv-Ethi-ZZ"},
    language_tag_expansion{"gn", "gn-Latn-PY"},
    language_tag_expansion{"gnd", "gnd-Latn-ZZ"},
    language_tag_expansion{"gng", "gng-Latn-ZZ"},
    language_tag_expansion{"god", "god-Latn-ZZ"},
    language_tag_expansion{"gof", "gof-Ethi-ZZ"},
    language_tag_expansion{"goi", "goi-Latn-ZZ"},
    language_tag_expansion{"gom", "gom-Deva-IN"},
    language_tag_expansion{"gon", "gon-Telu-IN"},
    language_tag_expansion{"gor", "gor-Latn-ID"},
    language_tag_expansion{"gos", "gos-Latn-NL"},
    language_tag_expansion{"got", "got-Goth-UA"},
    language_tag_expansion{"grb", "grb-Latn-ZZ"},
    language_tag_expansion{"grc", "grc-Cprt-CY"},
    language_tag_expansion{"grc-linb", "grc-Linb-GR"},
    language_tag_expansion{"grt", "grt-Beng-IN"},
    language_tag_expansion{"grw", "grw-Latn-ZZ"},
    language_tag_expansion{"gsw", "gsw-Latn-CH"},
    language_tag_expansion{"gu", "gu-Gujr-IN"},
    language_tag_expansion{"gub", "gub-Latn-BR"},
    language_tag_expansion{"guc", "guc-Latn-CO"},
    language_tag_expansion{"gud", "gud-Latn-ZZ"},
    language_tag_expansion{"gur", "gur-Latn-GH"},
    language_tag_expansion{"guw", "guw-Latn-ZZ"},
    language_tag_expansion{"gux", "gux-Latn-ZZ"},
    language_tag_expansion{"guz", "guz-Latn-KE"},
    language_tag_expansion{"gv", "gv-Latn-IM"},
    language_tag_expansion{"gvf", "gvf-Latn-ZZ"},
    language_tag_expansion{"gvr", "gvr-Deva-NP"},
    language_tag_expansion{"gvs", "gvs-Latn-ZZ"},
    language_tag_expansion{"gwc", "gwc-Arab-ZZ"},
    language_tag_expansion{"gwi", "gwi-Latn-CA"},
    language_tag_expansion{"gwt", "gwt-Arab-ZZ"},
    language_tag_expansion{"gyi", "gyi-Latn-ZZ"},
    language_tag_expansion{"ha", "ha-Latn-NG"},
    language_tag_expansion{"ha-cm", "ha-Arab-CM"},
    language_tag_expansion{"hag", "hag-Latn-ZZ"},
    language_tag_expansion{"hak", "hak-Hans-CN"},
    language_tag_expansion{"ham", "ham-Latn-ZZ"},
    language_tag_expansion{"ha-sd", "ha-Arab-SD"},
    language_tag_expansion{"haw", "haw-Latn-US"},
    language_tag_expansion{"haz", "haz-Arab-AF"},
    language_tag_expansion{"hbb", "hbb-Latn-ZZ"},
    language_tag_expansion{"hdy", "hdy-Ethi-ZZ"},
    language_tag_expansion{"he", "he-Hebr-IL"},
    language_tag_expansion{"hhy", "hhy-Latn-ZZ"},
    language_tag_expansion{"hi", "hi-Deva-IN"},
    language_tag_expansion{"hia", "hia-Latn-ZZ"},
    language_tag_expansion{"hif", "hif-Latn-FJ"},
    language_tag_expansion{"hig", "hig-Latn-ZZ"},
    language_tag_expansion{"hih", "hih-Latn-ZZ"},
    language_tag_expansion{"hil", "hil-Latn-PH"},
    language_tag_expansion{"hla", "hla-Latn-ZZ"},
    language_tag_expansion{"hlu", "hlu-Hluw-TR"},
    language_tag_expansion{"hmd", "hmd-Plrd-CN"},
    language_tag_expansion{"hmt", "hmt-Latn-ZZ"},
    language_tag_expansion{"hnd", "hnd-Arab-PK"},
    language_tag_expansion{"hne", "hne-Deva-IN"},
    language_tag_expansion{"hnj", "hnj-Hmnp-US"},
    language_tag_expansion{"hnn", "hnn-Latn-PH"},
    language_tag_expansion{"hno", "hno-Arab-PK"},
    language_tag_expansion{"ho", "ho-Latn-PG"},
    language_tag_expansion{"hoc", "hoc-Deva-IN"},
    language_tag_expansion{"hoj", "hoj-Deva-IN"},
    language_tag_expansion{"hot", "hot-Latn-ZZ"},
    language_tag_expansion{"hr", "hr-Latn-HR"},
    language_tag_expansion{"hsb", "hsb-Latn-DE"},
    language_tag_expansion{"hsn", "hsn-Hans-CN"},
    language_tag_expansion{"ht", "ht-Latn-HT"},
    language_tag_expansion{"hu", "hu-Latn-HU"},
    language_tag_expansion{"hui", "hui-Latn-ZZ"},
    language_tag_expansion{"hy", "hy-Armn-AM"},
    language_tag_expansion{"hz", "hz-Latn-NA"},
    language_tag_expansion{"ia", "ia-Latn-001"},
    language_tag_expansion{"i-ami", "ami"}, // grandfathered: not in ISO 639
    language_tag_expansion{"ian", "ian-Latn-ZZ"},
    language_tag_expansion{"iar", "iar-Latn-ZZ"},
    language_tag_expansion{"iba", "iba-Latn-MY"},
    language_tag_expansion{"ibb", "ibb-Latn-NG"},
    language_tag_expansion{"i-bnn", "bnn"}, // grandfathered: not in ISO 639
    language_tag_expansion{"iby", "iby-Latn-ZZ"},
    language_tag_expansion{"ica", "ica-Latn-ZZ"},
    language_tag_expansion{"ich", "ich-Latn-ZZ"},
    language_tag_expansion{"id", "id-Latn-ID"},
    language_tag_expansion{"idd", "idd-Latn-ZZ"},
    language_tag_expansion{"i-default", "en-Latn-US"}, // grandfathered
    language_tag_expansion{"idi", "idi-Latn-ZZ"},
    language_tag_expansion{"idu", "idu-Latn-ZZ"},
    language_tag_expansion{"ife", "ife-Latn-TG"},
    language_tag_expansion{"ig", "ig-Latn-NG"},
    language_tag_expansion{"igb", "igb-Latn-ZZ"},
    language_tag_expansion{"ige", "ige-Latn-ZZ"},
    language_tag_expansion{"i-hak", "hak-Hans-CN"}, // grandfathered
    language_tag_expansion{"ii", "ii-Yiii-CN"},
    language_tag_expansion{"ijj", "ijj-Latn-ZZ"},
    language_tag_expansion{"ik", "ik-Latn-US"},
    language_tag_expansion{"ikk", "ikk-Latn-ZZ"},
    language_tag_expansion{"i-klingon", "tlh"}, // grandfathered: not in ISO 639
    language_tag_expansion{"ikt", "ikt-Latn-CA"},
    language_tag_expansion{"ikw", "ikw-Latn-ZZ"},
    language_tag_expansion{"ikx", "ikx-Latn-ZZ"},
    language_tag_expansion{"ilo", "ilo-Latn-PH"},
    language_tag_expansion{"i-lux", "lb-Latn-LU"}, // grandfathered
    language_tag_expansion{"imo", "imo-Latn-ZZ"},
    language_tag_expansion{"in", "in-Latn-ID"},
    language_tag_expansion{"i-navajo", "nv-Latn-US"}, // grandfathered
    language_tag_expansion{"inh", "inh-Cyrl-RU"},
    language_tag_expansion{"io", "io-Latn-001"},
    language_tag_expansion{"iou", "iou-Latn-ZZ"},
    language_tag_expansion{"i-pwn", "pwn"}, // grandfathered: not in ISO 639
    language_tag_expansion{"iri", "iri-Latn-ZZ"},
    language_tag_expansion{"is", "is-Latn-IS"},
    language_tag_expansion{"it", "it-Latn-IT"},
    language_tag_expansion{"i-tao", "tao"}, // grandfathered: not in ISO 639
    language_tag_expansion{"i-tay", "tay"}, // grandfathered: not in ISO 639
    language_tag_expansion{"i-tsu", "tsu"}, // grandfathered: not in ISO 639
    language_tag_expansion{"iu", "iu-Cans-CA"},
    language_tag_expansion{"iw", "iw-Hebr-IL"},
    language_tag_expansion{"iwm", "iwm-Latn-ZZ"},
    language_tag_expansion{"iws", "iws-Latn-ZZ"},
    language_tag_expansion{"izh", "izh-Latn-RU"},
    language_tag_expansion{"izi", "izi-Latn-ZZ"},
    language_tag_expansion{"ja", "ja-Jpan-JP"},
    language_tag_expansion{"jab", "jab-Latn-ZZ"},
    language_tag_expansion{"jam", "jam-Latn-JM"},
    language_tag_expansion{"jar", "jar-Latn-ZZ"},
    language_tag_expansion{"jbo", "jbo-Latn-001"},
    language_tag_expansion{"jbu", "jbu-Latn-ZZ"},
    language_tag_expansion{"jen", "jen-Latn-ZZ"},
    language_tag_expansion{"jgk", "jgk-Latn-ZZ"},
    language_tag_expansion{"jgo", "jgo-Latn-CM"},
    language_tag_expansion{"ji", "ji-Hebr-UA"},
    language_tag_expansion{"jib", "jib-Latn-ZZ"},
    language_tag_expansion{"jmc", "jmc-Latn-TZ"},
    language_tag_expansion{"jml", "jml-Deva-NP"},
    language_tag_expansion{"jra", "jra-Latn-ZZ"},
    language_tag_expansion{"jut", "jut-Latn-DK"},
    language_tag_expansion{"jv", "jv-Latn-ID"},
    language_tag_expansion{"jw", "jw-Latn-ID"},
    language_tag_expansion{"ka", "ka-Geor-GE"},
    language_tag_expansion{"kaa", "kaa-Cyrl-UZ"},
    language_tag_expansion{"kab", "kab-Latn-DZ"},
    language_tag_expansion{"kac", "kac-Latn-MM"},
    language_tag_expansion{"kad", "kad-Latn-ZZ"},
    language_tag_expansion{"kai", "kai-Latn-ZZ"},
    language_tag_expansion{"kaj", "kaj-Latn-NG"},
    language_tag_expansion{"kam", "kam-Latn-KE"},
    language_tag_expansion{"kao", "kao-Latn-ML"},
    language_tag_expansion{"kbd", "kbd-Cyrl-RU"},
    language_tag_expansion{"kbm", "kbm-Latn-ZZ"},
    language_tag_expansion{"kbp", "kbp-Latn-ZZ"},
    language_tag_expansion{"kbq", "kbq-Latn-ZZ"},
    language_tag_expansion{"kbx", "kbx-Latn-ZZ"},
    language_tag_expansion{"kby", "kby-Arab-NE"},
    language_tag_expansion{"kcg", "kcg-Latn-NG"},
    language_tag_expansion{"kck", "kck-Latn-ZW"},
    language_tag_expansion{"kcl", "kcl-Latn-ZZ"},
    language_tag_expansion{"kct", "kct-Latn-ZZ"},
    language_tag_expansion{"kde", "kde-Latn-TZ"},
    language_tag_expansion{"kdh", "kdh-Latn-TG"},
    language_tag_expansion{"kdl", "kdl-Latn-ZZ"},
    language_tag_expansion{"kdt", "kdt-Thai-TH"},
    language_tag_expansion{"kea", "kea-Latn-CV"},
    language_tag_expansion{"ken", "ken-Latn-CM"},
    language_tag_expansion{"kez", "kez-Latn-ZZ"},
    language_tag_expansion{"kfo", "kfo-Latn-CI"},
    language_tag_expansion{"kfr", "kfr-Deva-IN"},
    language_tag_expansion{"kfy", "kfy-Deva-IN"},
    language_tag_expansion{"kg", "kg-Latn-CD"},
    language_tag_expansion{"kge", "kge-Latn-ID"},
    language_tag_expansion{"kgf", "kgf-Latn-ZZ"},
    language_tag_expansion{"kgp", "kgp-Latn-BR"},
    language_tag_expansion{"kha", "kha-Latn-IN"},
    language_tag_expansion{"khb", "khb-Talu-CN"},
    language_tag_expansion{"khn", "khn-Deva-IN"},
    language_tag_expansion{"khq", "khq-Latn-ML"},
    language_tag_expansion{"khs", "khs-Latn-ZZ"},
    language_tag_expansion{"kht", "kht-Mymr-IN"},
    language_tag_expansion{"khw", "khw-Arab-PK"},
    language_tag_expansion{"khz", "khz-Latn-ZZ"},
    language_tag_expansion{"ki", "ki-Latn-KE"},
    language_tag_expansion{"kij", "kij-Latn-ZZ"},
    language_tag_expansion{"kiu", "kiu-Latn-TR"},
    language_tag_expansion{"kiw", "kiw-Latn-ZZ"},
    language_tag_expansion{"kj", "kj-Latn-NA"},
    language_tag_expansion{"kjd", "kjd-Latn-ZZ"},
    language_tag_expansion{"kjg", "kjg-Laoo-LA"},
    language_tag_expansion{"kjs", "kjs-Latn-ZZ"},
    language_tag_expansion{"kjy", "kjy-Latn-ZZ"},
    language_tag_expansion{"kk", "kk-Cyrl-KZ"},
    language_tag_expansion{"kk-af", "kk-Arab-AF"},
    language_tag_expansion{"kk-arab", "kk-Arab-CN"},
    language_tag_expansion{"kkc", "kkc-Latn-ZZ"},
    language_tag_expansion{"kk-cn", "kk-Arab-CN"},
    language_tag_expansion{"kk-ir", "kk-Arab-IR"},
    language_tag_expansion{"kkj", "kkj-Latn-CM"},
    language_tag_expansion{"kk-mn", "kk-Arab-MN"},
    language_tag_expansion{"kl", "kl-Latn-GL"},
    language_tag_expansion{"kln", "kln-Latn-KE"},
    language_tag_expansion{"klq", "klq-Latn-ZZ"},
    language_tag_expansion{"klt", "klt-Latn-ZZ"},
    language_tag_expansion{"klx", "klx-Latn-ZZ"},
    language_tag_expansion{"km", "km-Khmr-KH"},
    language_tag_expansion{"kmb", "kmb-Latn-AO"},
    language_tag_expansion{"kmh", "kmh-Latn-ZZ"},
    language_tag_expansion{"kmo", "kmo-Latn-ZZ"},
    language_tag_expansion{"kms", "kms-Latn-ZZ"},
    language_tag_expansion{"kmu", "kmu-Latn-ZZ"},
    language_tag_expansion{"kmw", "kmw-Latn-ZZ"},
    language_tag_expansion{"kn", "kn-Knda-IN"},
    language_tag_expansion{"knf", "knf-Latn-GW"},
    language_tag_expansion{"knp", "knp-Latn-ZZ"},
    language_tag_expansion{"ko", "ko-Kore-KR"},
    language_tag_expansion{"koi", "koi-Cyrl-RU"},
    language_tag_expansion{"kok", "kok-Deva-IN"},
    language_tag_expansion{"kol", "kol-Latn-ZZ"},
    language_tag_expansion{"kos", "kos-Latn-FM"},
    language_tag_expansion{"koz", "koz-Latn-ZZ"},
    language_tag_expansion{"kpe", "kpe-Latn-LR"},
    language_tag_expansion{"kpf", "kpf-Latn-ZZ"},
    language_tag_expansion{"kpo", "kpo-Latn-ZZ"},
    language_tag_expansion{"kpr", "kpr-Latn-ZZ"},
    language_tag_expansion{"kpx", "kpx-Latn-ZZ"},
    language_tag_expansion{"kqb", "kqb-Latn-ZZ"},
    language_tag_expansion{"kqf", "kqf-Latn-ZZ"},
    language_tag_expansion{"kqs", "kqs-Latn-ZZ"},
    language_tag_expansion{"kqy", "kqy-Ethi-ZZ"},
    language_tag_expansion{"kr", "kr-Latn-ZZ"},
    language_tag_expansion{"krc", "krc-Cyrl-RU"},
    language_tag_expansion{"kri", "kri-Latn-SL"},
    language_tag_expansion{"krj", "krj-Latn-PH"},
    language_tag_expansion{"krl", "krl-Latn-RU"},
    language_tag_expansion{"krs", "krs-Latn-ZZ"},
    language_tag_expansion{"kru", "kru-Deva-IN"},
    language_tag_expansion{"ks", "ks-Arab-IN"},
    language_tag_expansion{"ksb", "ksb-Latn-TZ"},
    language_tag_expansion{"ksd", "ksd-Latn-ZZ"},
    language_tag_expansion{"ksf", "ksf-Latn-CM"},
    language_tag_expansion{"ksh", "ksh-Latn-DE"},
    language_tag_expansion{"ksj", "ksj-Latn-ZZ"},
    language_tag_expansion{"ksr", "ksr-Latn-ZZ"},
    language_tag_expansion{"ktb", "ktb-Ethi-ZZ"},
    language_tag_expansion{"ktm", "ktm-Latn-ZZ"},
    language_tag_expansion{"kto", "kto-Latn-ZZ"},
    language_tag_expansion{"ktr", "ktr-Latn-MY"},
    language_tag_expansion{"ku", "ku-Latn-TR"},
    language_tag_expansion{"ku-arab", "ku-Arab-IQ"},
    language_tag_expansion{"kub", "kub-Latn-ZZ"},
    language_tag_expansion{"kud", "kud-Latn-ZZ"},
    language_tag_expansion{"kue", "kue-Latn-ZZ"},
    language_tag_expansion{"kuj", "kuj-Latn-ZZ"},
    language_tag_expansion{"ku-lb", "ku-Arab-LB"},
    language_tag_expansion{"kum", "kum-Cyrl-RU"},
    language_tag_expansion{"kun", "kun-Latn-ZZ"},
    language_tag_expansion{"kup", "kup-Latn-ZZ"},
    language_tag_expansion{"kus", "kus-Latn-ZZ"},
    language_tag_expansion{"ku-yezi", "ku-Yezi-GE"},
    language_tag_expansion{"kv", "kv-Cyrl-RU"},
    language_tag_expansion{"kvg", "kvg-Latn-ZZ"},
    language_tag_expansion{"kvr", "kvr-Latn-ID"},
    language_tag_expansion{"kvx", "kvx-Arab-PK"},
    language_tag_expansion{"kw", "kw-Latn-GB"},
    language_tag_expansion{"kwj", "kwj-Latn-ZZ"},
    language_tag_expansion{"kwo", "kwo-Latn-ZZ"},
    language_tag_expansion{"kwq", "kwq-Latn-ZZ"},
    language_tag_expansion{"kxa", "kxa-Latn-ZZ"},
    language_tag_expansion{"kxc", "kxc-Ethi-ZZ"},
    language_tag_expansion{"kxe", "kxe-Latn-ZZ"},
    language_tag_expansion{"kxl", "kxl-Deva-IN"},
    language_tag_expansion{"kxm", "kxm-Thai-TH"},
    language_tag_expansion{"kxp", "kxp-Arab-PK"},
    language_tag_expansion{"kxw", "kxw-Latn-ZZ"},
    language_tag_expansion{"kxz", "kxz-Latn-ZZ"},
    language_tag_expansion{"ky", "ky-Cyrl-KG"},
    language_tag_expansion{"ky-arab", "ky-Arab-CN"},
    language_tag_expansion{"ky-cn", "ky-Arab-CN"},
    language_tag_expansion{"kye", "kye-Latn-ZZ"},
    language_tag_expansion{"ky-latn", "ky-Latn-TR"},
    language_tag_expansion{"ky-tr", "ky-Latn-TR"},
    language_tag_expansion{"kyx", "kyx-Latn-ZZ"},
    language_tag_expansion{"kzh", "kzh-Arab-ZZ"},
    language_tag_expansion{"kzj", "kzj-Latn-MY"},
    language_tag_expansion{"kzr", "kzr-Latn-ZZ"},
    language_tag_expansion{"kzt", "kzt-Latn-MY"},
    language_tag_expansion{"la", "la-Latn-VA"},
    language_tag_expansion{"lab", "lab-Lina-GR"},
    language_tag_expansion{"lad", "lad-Hebr-IL"},
    language_tag_expansion{"lag", "lag-Latn-TZ"},
    language_tag_expansion{"lah", "lah-Arab-PK"},
    language_tag_expansion{"laj", "laj-Latn-UG"},
    language_tag_expansion{"las", "las-Latn-ZZ"},
    language_tag_expansion{"lb", "lb-Latn-LU"},
    language_tag_expansion{"lbe", "lbe-Cyrl-RU"},
    language_tag_expansion{"lbu", "lbu-Latn-ZZ"},
    language_tag_expansion{"lbw", "lbw-Latn-ID"},
    language_tag_expansion{"lcm", "lcm-Latn-ZZ"},
    language_tag_expansion{"lcp", "lcp-Thai-CN"},
    language_tag_expansion{"ldb", "ldb-Latn-ZZ"},
    language_tag_expansion{"led", "led-Latn-ZZ"},
    language_tag_expansion{"lee", "lee-Latn-ZZ"},
    language_tag_expansion{"lem", "lem-Latn-ZZ"},
    language_tag_expansion{"lep", "lep-Lepc-IN"},
    language_tag_expansion{"leq", "leq-Latn-ZZ"},
    language_tag_expansion{"leu", "leu-Latn-ZZ"},
    language_tag_expansion{"lez", "lez-Cyrl-RU"},
    language_tag_expansion{"lg", "lg-Latn-UG"},
    language_tag_expansion{"lgg", "lgg-Latn-ZZ"},
    language_tag_expansion{"li", "li-Latn-NL"},
    language_tag_expansion{"lia", "lia-Latn-ZZ"},
    language_tag_expansion{"lid", "lid-Latn-ZZ"},
    language_tag_expansion{"lif", "lif-Deva-NP"},
    language_tag_expansion{"lif-limb", "lif-Limb-IN"},
    language_tag_expansion{"lig", "lig-Latn-ZZ"},
    language_tag_expansion{"lih", "lih-Latn-ZZ"},
    language_tag_expansion{"lij", "lij-Latn-IT"},
    language_tag_expansion{"lis", "lis-Lisu-CN"},
    language_tag_expansion{"ljp", "ljp-Latn-ID"},
    language_tag_expansion{"lki", "lki-Arab-IR"},
    language_tag_expansion{"lkt", "lkt-Latn-US"},
    language_tag_expansion{"lle", "lle-Latn-ZZ"},
    language_tag_expansion{"lln", "lln-Latn-ZZ"},
    language_tag_expansion{"lmn", "lmn-Telu-IN"},
    language_tag_expansion{"lmo", "lmo-Latn-IT"},
    language_tag_expansion{"lmp", "lmp-Latn-ZZ"},
    language_tag_expansion{"ln", "ln-Latn-CD"},
    language_tag_expansion{"lns", "lns-Latn-ZZ"},
    language_tag_expansion{"lnu", "lnu-Latn-ZZ"},
    language_tag_expansion{"lo", "lo-Laoo-LA"},
    language_tag_expansion{"loj", "loj-Latn-ZZ"},
    language_tag_expansion{"lok", "lok-Latn-ZZ"},
    language_tag_expansion{"lol", "lol-Latn-CD"},
    language_tag_expansion{"lor", "lor-Latn-ZZ"},
    language_tag_expansion{"los", "los-Latn-ZZ"},
    language_tag_expansion{"loz", "loz-Latn-ZM"},
    language_tag_expansion{"lrc", "lrc-Arab-IR"},
    language_tag_expansion{"lt", "lt-Latn-LT"},
    language_tag_expansion{"ltg", "ltg-Latn-LV"},
    language_tag_expansion{"lu", "lu-Latn-CD"},
    language_tag_expansion{"lua", "lua-Latn-CD"},
    language_tag_expansion{"luo", "luo-Latn-KE"},
    language_tag_expansion{"luy", "luy-Latn-KE"},
    language_tag_expansion{"luz", "luz-Arab-IR"},
    language_tag_expansion{"lv", "lv-Latn-LV"},
    language_tag_expansion{"lwl", "lwl-Thai-TH"},
    language_tag_expansion{"lzh", "lzh-Hans-CN"},
    language_tag_expansion{"lzz", "lzz-Latn-TR"},
    language_tag_expansion{"mad", "mad-Latn-ID"},
    language_tag_expansion{"maf", "maf-Latn-CM"},
    language_tag_expansion{"mag", "mag-Deva-IN"},
    language_tag_expansion{"mai", "mai-Deva-IN"},
    language_tag_expansion{"mak", "mak-Latn-ID"},
    language_tag_expansion{"man", "man-Latn-GM"},
    language_tag_expansion{"man-gn", "man-Nkoo-GN"},
    language_tag_expansion{"man-nkoo", "man-Nkoo-GN"},
    language_tag_expansion{"mas", "mas-Latn-KE"},
    language_tag_expansion{"maw", "maw-Latn-ZZ"},
    language_tag_expansion{"maz", "maz-Latn-MX"},
    language_tag_expansion{"mbh", "mbh-Latn-ZZ"},
    language_tag_expansion{"mbo", "mbo-Latn-ZZ"},
    language_tag_expansion{"mbq", "mbq-Latn-ZZ"},
    language_tag_expansion{"mbu", "mbu-Latn-ZZ"},
    language_tag_expansion{"mbw", "mbw-Latn-ZZ"},
    language_tag_expansion{"mci", "mci-Latn-ZZ"},
    language_tag_expansion{"mcp", "mcp-Latn-ZZ"},
    language_tag_expansion{"mcq", "mcq-Latn-ZZ"},
    language_tag_expansion{"mcr", "mcr-Latn-ZZ"},
    language_tag_expansion{"mcu", "mcu-Latn-ZZ"},
    language_tag_expansion{"mda", "mda-Latn-ZZ"},
    language_tag_expansion{"mde", "mde-Arab-ZZ"},
    language_tag_expansion{"mdf", "mdf-Cyrl-RU"},
    language_tag_expansion{"mdh", "mdh-Latn-PH"},
    language_tag_expansion{"mdj", "mdj-Latn-ZZ"},
    language_tag_expansion{"mdr", "mdr-Latn-ID"},
    language_tag_expansion{"mdx", "mdx-Ethi-ZZ"},
    language_tag_expansion{"med", "med-Latn-ZZ"},
    language_tag_expansion{"mee", "mee-Latn-ZZ"},
    language_tag_expansion{"mek", "mek-Latn-ZZ"},
    language_tag_expansion{"men", "men-Latn-SL"},
    language_tag_expansion{"mer", "mer-Latn-KE"},
    language_tag_expansion{"met", "met-Latn-ZZ"},
    language_tag_expansion{"meu", "meu-Latn-ZZ"},
    language_tag_expansion{"mfa", "mfa-Arab-TH"},
    language_tag_expansion{"mfe", "mfe-Latn-MU"},
    language_tag_expansion{"mfn", "mfn-Latn-ZZ"},
    language_tag_expansion{"mfo", "mfo-Latn-ZZ"},
    language_tag_expansion{"mfq", "mfq-Latn-ZZ"},
    language_tag_expansion{"mg", "mg-Latn-MG"},
    language_tag_expansion{"mgh", "mgh-Latn-MZ"},
    language_tag_expansion{"mgl", "mgl-Latn-ZZ"},
    language_tag_expansion{"mgo", "mgo-Latn-CM"},
    language_tag_expansion{"mgp", "mgp-Deva-NP"},
    language_tag_expansion{"mgy", "mgy-Latn-TZ"},
    language_tag_expansion{"mh", "mh-Latn-MH"},
    language_tag_expansion{"mhi", "mhi-Latn-ZZ"},
    language_tag_expansion{"mhl", "mhl-Latn-ZZ"},
    language_tag_expansion{"mi", "mi-Latn-NZ"},
    language_tag_expansion{"mif", "mif-Latn-ZZ"},
    language_tag_expansion{"min", "min-Latn-ID"},
    language_tag_expansion{"miw", "miw-Latn-ZZ"},
    language_tag_expansion{"mk", "mk-Cyrl-MK"},
    language_tag_expansion{"mki", "mki-Arab-ZZ"},
    language_tag_expansion{"mkl", "mkl-Latn-ZZ"},
    language_tag_expansion{"mkp", "mkp-Latn-ZZ"},
    language_tag_expansion{"mkw", "mkw-Latn-ZZ"},
    language_tag_expansion{"ml", "ml-Mlym-IN"},
    language_tag_expansion{"mle", "mle-Latn-ZZ"},
    language_tag_expansion{"mlp", "mlp-Latn-ZZ"},
    language_tag_expansion{"mls", "mls-Latn-SD"},
    language_tag_expansion{"mmo", "mmo-Latn-ZZ"},
    language_tag_expansion{"mmu", "mmu-Latn-ZZ"},
    language_tag_expansion{"mmx", "mmx-Latn-ZZ"},
    language_tag_expansion{"mn", "mn-Cyrl-MN"},
    language_tag_expansion{"mna", "mna-Latn-ZZ"},
    language_tag_expansion{"mn-cn", "mn-Mong-CN"},
    language_tag_expansion{"mnf", "mnf-Latn-ZZ"},
    language_tag_expansion{"mni", "mni-Beng-IN"},
    language_tag_expansion{"mn-mong", "mn-Mong-CN"},
    language_tag_expansion{"mnw", "mnw-Mymr-MM"},
    language_tag_expansion{"mo", "mo-Latn-RO"},
    language_tag_expansion{"moa", "moa-Latn-ZZ"},
    language_tag_expansion{"moe", "moe-Latn-CA"},
    language_tag_expansion{"moh", "moh-Latn-CA"},
    language_tag_expansion{"mos", "mos-Latn-BF"},
    language_tag_expansion{"mox", "mox-Latn-ZZ"},
    language_tag_expansion{"mpp", "mpp-Latn-ZZ"},
    language_tag_expansion{"mps", "mps-Latn-ZZ"},
    language_tag_expansion{"mpt", "mpt-Latn-ZZ"},
    language_tag_expansion{"mpx", "mpx-Latn-ZZ"},
    language_tag_expansion{"mql", "mql-Latn-ZZ"},
    language_tag_expansion{"mr", "mr-Deva-IN"},
    language_tag_expansion{"mrd", "mrd-Deva-NP"},
    language_tag_expansion{"mrj", "mrj-Cyrl-RU"},
    language_tag_expansion{"mro", "mro-Mroo-BD"},
    language_tag_expansion{"ms", "ms-Latn-MY"},
    language_tag_expansion{"ms-cc", "ms-Arab-CC"},
    language_tag_expansion{"mt", "mt-Latn-MT"},
    language_tag_expansion{"mtc", "mtc-Latn-ZZ"},
    language_tag_expansion{"mtf", "mtf-Latn-ZZ"},
    language_tag_expansion{"mti", "mti-Latn-ZZ"},
    language_tag_expansion{"mtr", "mtr-Deva-IN"},
    language_tag_expansion{"mua", "mua-Latn-CM"},
    language_tag_expansion{"mur", "mur-Latn-ZZ"},
    language_tag_expansion{"mus", "mus-Latn-US"},
    language_tag_expansion{"mva", "mva-Latn-ZZ"},
    language_tag_expansion{"mvn", "mvn-Latn-ZZ"},
    language_tag_expansion{"mvy", "mvy-Arab-PK"},
    language_tag_expansion{"mwk", "mwk-Latn-ML"},
    language_tag_expansion{"mwr", "mwr-Deva-IN"},
    language_tag_expansion{"mwv", "mwv-Latn-ID"},
    language_tag_expansion{"mww", "mww-Hmnp-US"},
    language_tag_expansion{"mxc", "mxc-Latn-ZW"},
    language_tag_expansion{"mxm", "mxm-Latn-ZZ"},
    language_tag_expansion{"my", "my-Mymr-MM"},
    language_tag_expansion{"myk", "myk-Latn-ZZ"},
    language_tag_expansion{"mym", "mym-Ethi-ZZ"},
    language_tag_expansion{"myv", "myv-Cyrl-RU"},
    language_tag_expansion{"myw", "myw-Latn-ZZ"},
    language_tag_expansion{"myx", "myx-Latn-UG"},
    language_tag_expansion{"myz", "myz-Mand-IR"},
    language_tag_expansion{"mzk", "mzk-Latn-ZZ"},
    language_tag_expansion{"mzm", "mzm-Latn-ZZ"},
    language_tag_expansion{"mzn", "mzn-Arab-IR"},
    language_tag_expansion{"mzp", "mzp-Latn-ZZ"},
    language_tag_expansion{"mzw", "mzw-Latn-ZZ"},
    language_tag_expansion{"mzz", "mzz-Latn-ZZ"},
    language_tag_expansion{"na", "na-Latn-NR"},
    language_tag_expansion{"nac", "nac-Latn-ZZ"},
    language_tag_expansion{"naf", "naf-Latn-ZZ"},
    language_tag_expansion{"nak", "nak-Latn-ZZ"},
    language_tag_expansion{"nan", "nan-Hans-CN"},
    language_tag_expansion{"nap", "nap-Latn-IT"},
    language_tag_expansion{"naq", "naq-Latn-NA"},
    language_tag_expansion{"nas", "nas-Latn-ZZ"},
    language_tag_expansion{"nb", "nb-Latn-NO"},
    language_tag_expansion{"nca", "nca-Latn-ZZ"},
    language_tag_expansion{"nce", "nce-Latn-ZZ"},
    language_tag_expansion{"ncf", "ncf-Latn-ZZ"},
    language_tag_expansion{"nch", "nch-Latn-MX"},
    language_tag_expansion{"nco", "nco-Latn-ZZ"},
    language_tag_expansion{"ncu", "ncu-Latn-ZZ"},
    language_tag_expansion{"nd", "nd-Latn-ZW"},
    language_tag_expansion{"ndc", "ndc-Latn-MZ"},
    language_tag_expansion{"nds", "nds-Latn-DE"},
    language_tag_expansion{"ne", "ne-Deva-NP"},
    language_tag_expansion{"neb", "neb-Latn-ZZ"},
    language_tag_expansion{"new", "new-Deva-NP"},
    language_tag_expansion{"nex", "nex-Latn-ZZ"},
    language_tag_expansion{"nfr", "nfr-Latn-ZZ"},
    language_tag_expansion{"ng", "ng-Latn-NA"},
    language_tag_expansion{"nga", "nga-Latn-ZZ"},
    language_tag_expansion{"ngb", "ngb-Latn-ZZ"},
    language_tag_expansion{"ngl", "ngl-Latn-MZ"},
    language_tag_expansion{"nhb", "nhb-Latn-ZZ"},
    language_tag_expansion{"nhe", "nhe-Latn-MX"},
    language_tag_expansion{"nhw", "nhw-Latn-MX"},
    language_tag_expansion{"nif", "nif-Latn-ZZ"},
    language_tag_expansion{"nii", "nii-Latn-ZZ"},
    language_tag_expansion{"nij", "nij-Latn-ID"},
    language_tag_expansion{"nin", "nin-Latn-ZZ"},
    language_tag_expansion{"niu", "niu-Latn-NU"},
    language_tag_expansion{"niy", "niy-Latn-ZZ"},
    language_tag_expansion{"niz", "niz-Latn-ZZ"},
    language_tag_expansion{"njo", "njo-Latn-IN"},
    language_tag_expansion{"nkg", "nkg-Latn-ZZ"},
    language_tag_expansion{"nko", "nko-Latn-ZZ"},
    language_tag_expansion{"nl", "nl-Latn-NL"},
    language_tag_expansion{"nmg", "nmg-Latn-CM"},
    language_tag_expansion{"nmz", "nmz-Latn-ZZ"},
    language_tag_expansion{"nn", "nn-Latn-NO"},
    language_tag_expansion{"nnf", "nnf-Latn-ZZ"},
    language_tag_expansion{"nnh", "nnh-Latn-CM"},
    language_tag_expansion{"nnk", "nnk-Latn-ZZ"},
    language_tag_expansion{"nnm", "nnm-Latn-ZZ"},
    language_tag_expansion{"nnp", "nnp-Wcho-IN"},
    language_tag_expansion{"no", "no-Latn-NO"},
    language_tag_expansion{"no-bok", "nb-Latn-NO"}, // grandfathered
    language_tag_expansion{"nod", "nod-Lana-TH"},
    language_tag_expansion{"noe", "noe-Deva-IN"},
    language_tag_expansion{"non", "non-Runr-SE"},
    language_tag_expansion{"no-nyn", "nn-Latn-NO"}, // grandfathered
    language_tag_expansion{"nop", "nop-Latn-ZZ"},
    language_tag_expansion{"nou", "nou-Latn-ZZ"},
    language_tag_expansion{"nqo", "nqo-Nkoo-GN"},
    language_tag_expansion{"nr", "nr-Latn-ZA"},
    language_tag_expansion{"nrb", "nrb-Latn-ZZ"},
    language_tag_expansion{"nsk", "nsk-Cans-CA"},
    language_tag_expansion{"nsn", "nsn-Latn-ZZ"},
    language_tag_expansion{"nso", "nso-Latn-ZA"},
    language_tag_expansion{"nss", "nss-Latn-ZZ"},
    language_tag_expansion{"nst", "nst-Tnsa-IN"},
    language_tag_expansion{"ntm", "ntm-Latn-ZZ"},
    language_tag_expansion{"ntr", "ntr-Latn-ZZ"},
    language_tag_expansion{"nui", "nui-Latn-ZZ"},
    language_tag_expansion{"nup", "nup-Latn-ZZ"},
    language_tag_expansion{"nus", "nus-Latn-SS"},
    language_tag_expansion{"nuv", "nuv-Latn-ZZ"},
    language_tag_expansion{"nux", "nux-Latn-ZZ"},
    language_tag_expansion{"nv", "nv-Latn-US"},
    language_tag_expansion{"nwb", "nwb-Latn-ZZ"},
    language_tag_expansion{"nxq", "nxq-Latn-CN"},
    language_tag_expansion{"nxr", "nxr-Latn-ZZ"},
    language_tag_expansion{"ny", "ny-Latn-MW"},
    language_tag_expansion{"nym", "nym-Latn-TZ"},
    language_tag_expansion{"nyn", "nyn-Latn-UG"},
    language_tag_expansion{"nzi", "nzi-Latn-GH"},
    language_tag_expansion{"oc", "oc-Latn-FR"},
    language_tag_expansion{"ogc", "ogc-Latn-ZZ"},
    language_tag_expansion{"okr", "okr-Latn-ZZ"},
    language_tag_expansion{"okv", "okv-Latn-ZZ"},
    language_tag_expansion{"om", "om-Latn-ET"},
    language_tag_expansion{"ong", "ong-Latn-ZZ"},
    language_tag_expansion{"onn", "onn-Latn-ZZ"},
    language_tag_expansion{"ons", "ons-Latn-ZZ"},
    language_tag_expansion{"opm", "opm-Latn-ZZ"},
    language_tag_expansion{"or", "or-Orya-IN"},
    language_tag_expansion{"oro", "oro-Latn-ZZ"},
    language_tag_expansion{"oru", "oru-Arab-ZZ"},
    language_tag_expansion{"os", "os-Cyrl-GE"},
    language_tag_expansion{"osa", "osa-Osge-US"},
    language_tag_expansion{"ota", "ota-Arab-ZZ"},
    language_tag_expansion{"otk", "otk-Orkh-MN"},
    language_tag_expansion{"oui", "oui-Ougr-143"},
    language_tag_expansion{"ozm", "ozm-Latn-ZZ"},
    language_tag_expansion{"pa", "pa-Guru-IN"},
    language_tag_expansion{"pa-arab", "pa-Arab-PK"},
    language_tag_expansion{"pag", "pag-Latn-PH"},
    language_tag_expansion{"pal", "pal-Phli-IR"},
    language_tag_expansion{"pal-phlp", "pal-Phlp-CN"},
    language_tag_expansion{"pam", "pam-Latn-PH"},
    language_tag_expansion{"pap", "pap-Latn-AW"},
    language_tag_expansion{"pa-pk", "pa-Arab-PK"},
    language_tag_expansion{"pau", "pau-Latn-PW"},
    language_tag_expansion{"pbi", "pbi-Latn-ZZ"},
    language_tag_expansion{"pcd", "pcd-Latn-FR"},
    language_tag_expansion{"pcm", "pcm-Latn-NG"},
    language_tag_expansion{"pdc", "pdc-Latn-US"},
    language_tag_expansion{"pdt", "pdt-Latn-CA"},
    language_tag_expansion{"ped", "ped-Latn-ZZ"},
    language_tag_expansion{"peo", "peo-Xpeo-IR"},
    language_tag_expansion{"pex", "pex-Latn-ZZ"},
    language_tag_expansion{"pfl", "pfl-Latn-DE"},
    language_tag_expansion{"phl", "phl-Arab-ZZ"},
    language_tag_expansion{"phn", "phn-Phnx-LB"},
    language_tag_expansion{"pil", "pil-Latn-ZZ"},
    language_tag_expansion{"pip", "pip-Latn-ZZ"},
    language_tag_expansion{"pka", "pka-Brah-IN"},
    language_tag_expansion{"pko", "pko-Latn-KE"},
    language_tag_expansion{"pl", "pl-Latn-PL"},
    language_tag_expansion{"pla", "pla-Latn-ZZ"},
    language_tag_expansion{"pms", "pms-Latn-IT"},
    language_tag_expansion{"png", "png-Latn-ZZ"},
    language_tag_expansion{"pnn", "pnn-Latn-ZZ"},
    language_tag_expansion{"pnt", "pnt-Grek-GR"},
    language_tag_expansion{"pon", "pon-Latn-FM"},
    language_tag_expansion{"ppa", "ppa-Deva-IN"},
    language_tag_expansion{"ppo", "ppo-Latn-ZZ"},
    language_tag_expansion{"pra", "pra-Khar-PK"},
    language_tag_expansion{"prd", "prd-Arab-IR"},
    language_tag_expansion{"prg", "prg-Latn-001"},
    language_tag_expansion{"ps", "ps-Arab-AF"},
    language_tag_expansion{"pss", "pss-Latn-ZZ"},
    language_tag_expansion{"pt", "pt-Latn-BR"},
    language_tag_expansion{"ptp", "ptp-Latn-ZZ"},
    language_tag_expansion{"puu", "puu-Latn-GA"},
    language_tag_expansion{"pwa", "pwa-Latn-ZZ"},
    language_tag_expansion{"qu", "qu-Latn-PE"},
    language_tag_expansion{"quc", "quc-Latn-GT"},
    language_tag_expansion{"qug", "qug-Latn-EC"},
    language_tag_expansion{"rai", "rai-Latn-ZZ"},
    language_tag_expansion{"raj", "raj-Deva-IN"},
    language_tag_expansion{"rao", "rao-Latn-ZZ"},
    language_tag_expansion{"rcf", "rcf-Latn-RE"},
    language_tag_expansion{"rej", "rej-Latn-ID"},
    language_tag_expansion{"rel", "rel-Latn-ZZ"},
    language_tag_expansion{"res", "res-Latn-ZZ"},
    language_tag_expansion{"rgn", "rgn-Latn-IT"},
    language_tag_expansion{"rhg", "rhg-Rohg-MM"},
    language_tag_expansion{"ria", "ria-Latn-IN"},
    language_tag_expansion{"rif", "rif-Tfng-MA"},
    language_tag_expansion{"rif-nl", "rif-Latn-NL"},
    language_tag_expansion{"rjs", "rjs-Deva-NP"},
    language_tag_expansion{"rkt", "rkt-Beng-BD"},
    language_tag_expansion{"rm", "rm-Latn-CH"},
    language_tag_expansion{"rmf", "rmf-Latn-FI"},
    language_tag_expansion{"rmo", "rmo-Latn-CH"},
    language_tag_expansion{"rmt", "rmt-Arab-IR"},
    language_tag_expansion{"rmu", "rmu-Latn-SE"},
    language_tag_expansion{"rn", "rn-Latn-BI"},
    language_tag_expansion{"rna", "rna-Latn-ZZ"},
    language_tag_expansion{"rng", "rng-Latn-MZ"},
    language_tag_expansion{"ro", "ro-Latn-RO"},
    language_tag_expansion{"rob", "rob-Latn-ID"},
    language_tag_expansion{"rof", "rof-Latn-TZ"},
    language_tag_expansion{"roo", "roo-Latn-ZZ"},
    language_tag_expansion{"rro", "rro-Latn-ZZ"},
    language_tag_expansion{"rtm", "rtm-Latn-FJ"},
    language_tag_expansion{"ru", "ru-Cyrl-RU"},
    language_tag_expansion{"rue", "rue-Cyrl-UA"},
    language_tag_expansion{"rug", "rug-Latn-SB"},
    language_tag_expansion{"rw", "rw-Latn-RW"},
    language_tag_expansion{"rwk", "rwk-Latn-TZ"},
    language_tag_expansion{"rwo", "rwo-Latn-ZZ"},
    language_tag_expansion{"ryu", "ryu-Kana-JP"},
    language_tag_expansion{"sa", "sa-Deva-IN"},
    language_tag_expansion{"saf", "saf-Latn-GH"},
    language_tag_expansion{"sah", "sah-Cyrl-RU"},
    language_tag_expansion{"saq", "saq-Latn-KE"},
    language_tag_expansion{"sas", "sas-Latn-ID"},
    language_tag_expansion{"sat", "sat-Olck-IN"},
    language_tag_expansion{"sav", "sav-Latn-SN"},
    language_tag_expansion{"saz", "saz-Saur-IN"},
    language_tag_expansion{"sba", "sba-Latn-ZZ"},
    language_tag_expansion{"sbe", "sbe-Latn-ZZ"},
    language_tag_expansion{"sbp", "sbp-Latn-TZ"},
    language_tag_expansion{"sc", "sc-Latn-IT"},
    language_tag_expansion{"sck", "sck-Deva-IN"},
    language_tag_expansion{"scl", "scl-Arab-ZZ"},
    language_tag_expansion{"scn", "scn-Latn-IT"},
    language_tag_expansion{"sco", "sco-Latn-GB"},
    language_tag_expansion{"scs", "scs-Latn-CA"},
    language_tag_expansion{"sd", "sd-Arab-PK"},
    language_tag_expansion{"sdc", "sdc-Latn-IT"},
    language_tag_expansion{"sd-deva", "sd-Deva-IN"},
    language_tag_expansion{"sd-dind", "sd-Sind-IN"},
    language_tag_expansion{"sdh", "sdh-Arab-IR"},
    language_tag_expansion{"sd-khoj", "sd-Khoj-IN"},
    language_tag_expansion{"se", "se-Latn-NO"},
    language_tag_expansion{"sef", "sef-Latn-CI"},
    language_tag_expansion{"seh", "seh-Latn-MZ"},
    language_tag_expansion{"sei", "sei-Latn-MX"},
    language_tag_expansion{"ses", "ses-Latn-ML"},
    language_tag_expansion{"sg", "sg-Latn-CF"},
    language_tag_expansion{"sga", "sga-Ogam-IE"},
    language_tag_expansion{"sgn-be-fr", "fr-Sigw-BE"}, // grandfathered
    language_tag_expansion{"sgn-be-nl", "nl-Sigw-BE"}, // grandfathered
    language_tag_expansion{"sgn-ch-de", "de-Sigw-CH"}, // grandfathered
    language_tag_expansion{"sgs", "sgs-Latn-LT"},
    language_tag_expansion{"sgw", "sgw-Ethi-ZZ"},
    language_tag_expansion{"sgz", "sgz-Latn-ZZ"},
    language_tag_expansion{"shi", "shi-Tfng-MA"},
    language_tag_expansion{"shk", "shk-Latn-ZZ"},
    language_tag_expansion{"shn", "shn-Mymr-MM"},
    language_tag_expansion{"shu", "shu-Arab-ZZ"},
    language_tag_expansion{"si", "si-Sinh-LK"},
    language_tag_expansion{"sid", "sid-Latn-ET"},
    language_tag_expansion{"sig", "sig-Latn-ZZ"},
    language_tag_expansion{"sil", "sil-Latn-ZZ"},
    language_tag_expansion{"sim", "sim-Latn-ZZ"},
    language_tag_expansion{"sjr", "sjr-Latn-ZZ"},
    language_tag_expansion{"sk", "sk-Latn-SK"},
    language_tag_expansion{"skc", "skc-Latn-ZZ"},
    language_tag_expansion{"skr", "skr-Arab-PK"},
    language_tag_expansion{"sks", "sks-Latn-ZZ"},
    language_tag_expansion{"sl", "sl-Latn-SI"},
    language_tag_expansion{"sld", "sld-Latn-ZZ"},
    language_tag_expansion{"sli", "sli-Latn-PL"},
    language_tag_expansion{"sll", "sll-Latn-ZZ"},
    language_tag_expansion{"sly", "sly-Latn-ID"},
    language_tag_expansion{"sm", "sm-Latn-WS"},
    language_tag_expansion{"sma", "sma-Latn-SE"},
    language_tag_expansion{"smj", "smj-Latn-SE"},
    language_tag_expansion{"smn", "smn-Latn-FI"},
    language_tag_expansion{"smp", "smp-Samr-IL"},
    language_tag_expansion{"smq", "smq-Latn-ZZ"},
    language_tag_expansion{"sms", "sms-Latn-FI"},
    language_tag_expansion{"sn", "sn-Latn-ZW"},
    language_tag_expansion{"snc", "snc-Latn-ZZ"},
    language_tag_expansion{"snk", "snk-Latn-ML"},
    language_tag_expansion{"snp", "snp-Latn-ZZ"},
    language_tag_expansion{"snx", "snx-Latn-ZZ"},
    language_tag_expansion{"sny", "sny-Latn-ZZ"},
    language_tag_expansion{"so", "so-Latn-SO"},
    language_tag_expansion{"sog", "sog-Sogd-UZ"},
    language_tag_expansion{"sok", "sok-Latn-ZZ"},
    language_tag_expansion{"soq", "soq-Latn-ZZ"},
    language_tag_expansion{"sou", "sou-Thai-TH"},
    language_tag_expansion{"soy", "soy-Latn-ZZ"},
    language_tag_expansion{"spd", "spd-Latn-ZZ"},
    language_tag_expansion{"spl", "spl-Latn-ZZ"},
    language_tag_expansion{"sps", "sps-Latn-ZZ"},
    language_tag_expansion{"sq", "sq-Latn-AL"},
    language_tag_expansion{"sr", "sr-Cyrl-RS"},
    language_tag_expansion{"srb", "srb-Sora-IN"},
    language_tag_expansion{"sr-me", "sr-Latn-ME"},
    language_tag_expansion{"srn", "srn-Latn-SR"},
    language_tag_expansion{"srr", "srr-Latn-SN"},
    language_tag_expansion{"sr-ro", "sr-Latn-RO"},
    language_tag_expansion{"sr-ru", "sr-Latn-RU"},
    language_tag_expansion{"sr-tr", "sr-Latn-TR"},
    language_tag_expansion{"srx", "srx-Deva-IN"},
    language_tag_expansion{"ss", "ss-Latn-ZA"},
    language_tag_expansion{"ssd", "ssd-Latn-ZZ"},
    language_tag_expansion{"ssg", "ssg-Latn-ZZ"},
    language_tag_expansion{"ssy", "ssy-Latn-ER"},
    language_tag_expansion{"st", "st-Latn-ZA"},
    language_tag_expansion{"stk", "stk-Latn-ZZ"},
    language_tag_expansion{"stq", "stq-Latn-DE"},
    language_tag_expansion{"su", "su-Latn-ID"},
    language_tag_expansion{"sua", "sua-Latn-ZZ"},
    language_tag_expansion{"sue", "sue-Latn-ZZ"},
    language_tag_expansion{"suk", "suk-Latn-TZ"},
    language_tag_expansion{"sur", "sur-Latn-ZZ"},
    language_tag_expansion{"sus", "sus-Latn-GN"},
    language_tag_expansion{"sv", "sv-Latn-SE"},
    language_tag_expansion{"sw", "sw-Latn-TZ"},
    language_tag_expansion{"swb", "swb-Arab-YT"},
    language_tag_expansion{"swc", "swc-Latn-CD"},
    language_tag_expansion{"swg", "swg-Latn-DE"},
    language_tag_expansion{"swp", "swp-Latn-ZZ"},
    language_tag_expansion{"swv", "swv-Deva-IN"},
    language_tag_expansion{"sxn", "sxn-Latn-ID"},
    language_tag_expansion{"sxw", "sxw-Latn-ZZ"},
    language_tag_expansion{"syl", "syl-Beng-BD"},
    language_tag_expansion{"syr", "syr-Syrc-IQ"},
    language_tag_expansion{"szl", "szl-Latn-PL"},
    language_tag_expansion{"ta", "ta-Taml-IN"},
    language_tag_expansion{"taj", "taj-Deva-NP"},
    language_tag_expansion{"tal", "tal-Latn-ZZ"},
    language_tag_expansion{"tan", "tan-Latn-ZZ"},
    language_tag_expansion{"taq", "taq-Latn-ZZ"},
    language_tag_expansion{"tbc", "tbc-Latn-ZZ"},
    language_tag_expansion{"tbd", "tbd-Latn-ZZ"},
    language_tag_expansion{"tbf", "tbf-Latn-ZZ"},
    language_tag_expansion{"tbg", "tbg-Latn-ZZ"},
    language_tag_expansion{"tbo", "tbo-Latn-ZZ"},
    language_tag_expansion{"tbw", "tbw-Latn-PH"},
    language_tag_expansion{"tbz", "tbz-Latn-ZZ"},
    language_tag_expansion{"tci", "tci-Latn-ZZ"},
    language_tag_expansion{"tcy", "tcy-Knda-IN"},
    language_tag_expansion{"tdd", "tdd-Tale-CN"},
    language_tag_expansion{"tdg", "tdg-Deva-NP"},
    language_tag_expansion{"tdh", "tdh-Deva-NP"},
    language_tag_expansion{"tdu", "tdu-Latn-MY"},
    language_tag_expansion{"te", "te-Telu-IN"},
    language_tag_expansion{"ted", "ted-Latn-ZZ"},
    language_tag_expansion{"tem", "tem-Latn-SL"},
    language_tag_expansion{"teo", "teo-Latn-UG"},
    language_tag_expansion{"tet", "tet-Latn-TL"},
    language_tag_expansion{"tfi", "tfi-Latn-ZZ"},
    language_tag_expansion{"tg", "tg-Cyrl-TJ"},
    language_tag_expansion{"tg-arab", "tg-Arab-PK"},
    language_tag_expansion{"tgc", "tgc-Latn-ZZ"},
    language_tag_expansion{"tgo", "tgo-Latn-ZZ"},
    language_tag_expansion{"tg-pk", "tg-Arab-PK"},
    language_tag_expansion{"tgu", "tgu-Latn-ZZ"},
    language_tag_expansion{"th", "th-Thai-TH"},
    language_tag_expansion{"thl", "thl-Deva-NP"},
    language_tag_expansion{"thq", "thq-Deva-NP"},
    language_tag_expansion{"thr", "thr-Deva-NP"},
    language_tag_expansion{"ti", "ti-Ethi-ET"},
    language_tag_expansion{"tif", "tif-Latn-ZZ"},
    language_tag_expansion{"tig", "tig-Ethi-ER"},
    language_tag_expansion{"tik", "tik-Latn-ZZ"},
    language_tag_expansion{"tim", "tim-Latn-ZZ"},
    language_tag_expansion{"tio", "tio-Latn-ZZ"},
    language_tag_expansion{"tiv", "tiv-Latn-NG"},
    language_tag_expansion{"tk", "tk-Latn-TM"},
    language_tag_expansion{"tkl", "tkl-Latn-TK"},
    language_tag_expansion{"tkr", "tkr-Latn-AZ"},
    language_tag_expansion{"tkt", "tkt-Deva-NP"},
    language_tag_expansion{"tl", "tl-Latn-PH"},
    language_tag_expansion{"tlf", "tlf-Latn-ZZ"},
    language_tag_expansion{"tlx", "tlx-Latn-ZZ"},
    language_tag_expansion{"tly", "tly-Latn-AZ"},
    language_tag_expansion{"tmh", "tmh-Latn-NE"},
    language_tag_expansion{"tmy", "tmy-Latn-ZZ"},
    language_tag_expansion{"tn", "tn-Latn-ZA"},
    language_tag_expansion{"tnh", "tnh-Latn-ZZ"},
    language_tag_expansion{"to", "to-Latn-TO"},
    language_tag_expansion{"tof", "tof-Latn-ZZ"},
    language_tag_expansion{"tog", "tog-Latn-MW"},
    language_tag_expansion{"toq", "toq-Latn-ZZ"},
    language_tag_expansion{"tpi", "tpi-Latn-PG"},
    language_tag_expansion{"tpm", "tpm-Latn-ZZ"},
    language_tag_expansion{"tpz", "tpz-Latn-ZZ"},
    language_tag_expansion{"tqo", "tqo-Latn-ZZ"},
    language_tag_expansion{"tr", "tr-Latn-TR"},
    language_tag_expansion{"tru", "tru-Latn-TR"},
    language_tag_expansion{"trv", "trv-Latn-TW"},
    language_tag_expansion{"trw", "trw-Arab-PK"},
    language_tag_expansion{"ts", "ts-Latn-ZA"},
    language_tag_expansion{"tsd", "tsd-Grek-GR"},
    language_tag_expansion{"tsf", "tsf-Deva-NP"},
    language_tag_expansion{"tsg", "tsg-Latn-PH"},
    language_tag_expansion{"tsj", "tsj-Tibt-BT"},
    language_tag_expansion{"tsw", "tsw-Latn-ZZ"},
    language_tag_expansion{"tt", "tt-Cyrl-RU"},
    language_tag_expansion{"ttd", "ttd-Latn-ZZ"},
    language_tag_expansion{"tte", "tte-Latn-ZZ"},
    language_tag_expansion{"ttj", "ttj-Latn-UG"},
    language_tag_expansion{"ttr", "ttr-Latn-ZZ"},
    language_tag_expansion{"tts", "tts-Thai-TH"},
    language_tag_expansion{"ttt", "ttt-Latn-AZ"},
    language_tag_expansion{"tuh", "tuh-Latn-ZZ"},
    language_tag_expansion{"tul", "tul-Latn-ZZ"},
    language_tag_expansion{"tum", "tum-Latn-MW"},
    language_tag_expansion{"tuq", "tuq-Latn-ZZ"},
    language_tag_expansion{"tvd", "tvd-Latn-ZZ"},
    language_tag_expansion{"tvl", "tvl-Latn-TV"},
    language_tag_expansion{"tvu", "tvu-Latn-ZZ"},
    language_tag_expansion{"twh", "twh-Latn-ZZ"},
    language_tag_expansion{"twq", "twq-Latn-NE"},
    language_tag_expansion{"txg", "txg-Tang-CN"},
    language_tag_expansion{"txo", "txo-Toto-IN"},
    language_tag_expansion{"ty", "ty-Latn-PF"},
    language_tag_expansion{"tya", "tya-Latn-ZZ"},
    language_tag_expansion{"tyv", "tyv-Cyrl-RU"},
    language_tag_expansion{"tzm", "tzm-Latn-MA"},
    language_tag_expansion{"ubu", "ubu-Latn-ZZ"},
    language_tag_expansion{"udi", "udi-Aghb-RU"},
    language_tag_expansion{"udm", "udm-Cyrl-RU"},
    language_tag_expansion{"ug", "ug-Arab-CN"},
    language_tag_expansion{"uga", "uga-Ugar-SY"},
    language_tag_expansion{"ug-cyrl", "ug-Cyrl-KZ"},
    language_tag_expansion{"ug-kz", "ug-Cyrl-KZ"},
    language_tag_expansion{"ug-mn", "ug-Cyrl-MN"},
    language_tag_expansion{"uk", "uk-Cyrl-UA"},
    language_tag_expansion{"uli", "uli-Latn-FM"},
    language_tag_expansion{"umb", "umb-Latn-AO"},
    language_tag_expansion{"und", "en-Latn-US"},
    language_tag_expansion{"und-002", "en-Latn-NG"},
    language_tag_expansion{"und-003", "en-Latn-US"},
    language_tag_expansion{"und-005", "pt-Latn-BR"},
    language_tag_expansion{"und-009", "en-Latn-AU"},
    language_tag_expansion{"und-011", "en-Latn-NG"},
    language_tag_expansion{"und-013", "es-Latn-MX"},
    language_tag_expansion{"und-014", "sw-Latn-TZ"},
    language_tag_expansion{"und-015", "ar-Arab-EG"},
    language_tag_expansion{"und-017", "sw-Latn-CD"},
    language_tag_expansion{"und-018", "en-Latn-ZA"},
    language_tag_expansion{"und-019", "en-Latn-US"},
    language_tag_expansion{"und-021", "en-Latn-US"},
    language_tag_expansion{"und-029", "es-Latn-CU"},
    language_tag_expansion{"und-030", "zh-Hans-CN"},
    language_tag_expansion{"und-034", "hi-Deva-IN"},
    language_tag_expansion{"und-035", "id-Latn-ID"},
    language_tag_expansion{"und-039", "it-Latn-IT"},
    language_tag_expansion{"und-053", "en-Latn-AU"},
    language_tag_expansion{"und-054", "en-Latn-PG"},
    language_tag_expansion{"und-057", "en-Latn-GU"},
    language_tag_expansion{"und-061", "sm-Latn-WS"},
    language_tag_expansion{"und-142", "zh-Hans-CN"},
    language_tag_expansion{"und-143", "uz-Latn-UZ"},
    language_tag_expansion{"und-145", "ar-Arab-SA"},
    language_tag_expansion{"und-150", "ru-Cyrl-RU"},
    language_tag_expansion{"und-151", "ru-Cyrl-RU"},
    language_tag_expansion{"und-154", "en-Latn-GB"},
    language_tag_expansion{"und-155", "de-Latn-DE"},
    language_tag_expansion{"und-202", "en-Latn-NG"},
    language_tag_expansion{"und-419", "es-Latn-419"},
    language_tag_expansion{"und-ad", "ca-Latn-AD"},
    language_tag_expansion{"und-adlm", "ff-Adlm-GN"},
    language_tag_expansion{"und-ae", "ar-Arab-AE"},
    language_tag_expansion{"und-af", "fa-Arab-AF"},
    language_tag_expansion{"und-aghb", "udi-Aghb-RU"},
    language_tag_expansion{"und-ahom", "aho-Ahom-IN"},
    language_tag_expansion{"und-al", "sq-Latn-AL"},
    language_tag_expansion{"und-am", "hy-Armn-AM"},
    language_tag_expansion{"und-ao", "pt-Latn-AO"},
    language_tag_expansion{"und-aq", "und-Latn-AQ"},
    language_tag_expansion{"und-ar", "es-Latn-AR"},
    language_tag_expansion{"und-arab", "ar-Arab-EG"},
    language_tag_expansion{"und-arab-cc", "ms-Arab-CC"},
    language_tag_expansion{"und-arab-cn", "ug-Arab-CN"},
    language_tag_expansion{"und-arab-gb", "ks-Arab-GB"},
    language_tag_expansion{"und-arab-id", "ms-Arab-ID"},
    language_tag_expansion{"und-arab-in", "ur-Arab-IN"},
    language_tag_expansion{"und-arab-kh", "cja-Arab-KH"},
    language_tag_expansion{"und-arab-mm", "rhg-Arab-MM"},
    language_tag_expansion{"und-arab-mn", "kk-Arab-MN"},
    language_tag_expansion{"und-arab-mu", "ur-Arab-MU"},
    language_tag_expansion{"und-arab-ng", "ha-Arab-NG"},
    language_tag_expansion{"und-arab-pk", "ur-Arab-PK"},
    language_tag_expansion{"und-arab-tg", "apd-Arab-TG"},
    language_tag_expansion{"und-arab-th", "mfa-Arab-TH"},
    language_tag_expansion{"und-arab-tj", "fa-Arab-TJ"},
    language_tag_expansion{"und-arab-tr", "az-Arab-TR"},
    language_tag_expansion{"und-arab-yt", "swb-Arab-YT"},
    language_tag_expansion{"und-armi", "arc-Armi-IR"},
    language_tag_expansion{"und-armn", "hy-Armn-AM"},
    language_tag_expansion{"und-as", "sm-Latn-AS"},
    language_tag_expansion{"und-at", "de-Latn-AT"},
    language_tag_expansion{"und-avst", "ae-Avst-IR"},
    language_tag_expansion{"und-aw", "nl-Latn-AW"},
    language_tag_expansion{"und-ax", "sv-Latn-AX"},
    language_tag_expansion{"und-az", "az-Latn-AZ"},
    language_tag_expansion{"und-ba", "bs-Latn-BA"},
    language_tag_expansion{"und-bali", "ban-Bali-ID"},
    language_tag_expansion{"und-bamu", "bax-Bamu-CM"},
    language_tag_expansion{"und-bass", "bsq-Bass-LR"},
    language_tag_expansion{"und-batk", "bbc-Batk-ID"},
    language_tag_expansion{"und-bd", "bn-Beng-BD"},
    language_tag_expansion{"und-be", "nl-Latn-BE"},
    language_tag_expansion{"und-beng", "bn-Beng-BD"},
    language_tag_expansion{"und-bf", "fr-Latn-BF"},
    language_tag_expansion{"und-bg", "bg-Cyrl-BG"},
    language_tag_expansion{"und-bh", "ar-Arab-BH"},
    language_tag_expansion{"und-bhks", "sa-Bhks-IN"},
    language_tag_expansion{"und-bi", "rn-Latn-BI"},
    language_tag_expansion{"und-bj", "fr-Latn-BJ"},
    language_tag_expansion{"und-bl", "fr-Latn-BL"},
    language_tag_expansion{"und-bn", "ms-Latn-BN"},
    language_tag_expansion{"und-bo", "es-Latn-BO"},
    language_tag_expansion{"und-bopo", "zh-Bopo-TW"},
    language_tag_expansion{"und-bq", "pap-Latn-BQ"},
    language_tag_expansion{"und-br", "pt-Latn-BR"},
    language_tag_expansion{"und-brah", "pka-Brah-IN"},
    language_tag_expansion{"und-brai", "fr-Brai-FR"},
    language_tag_expansion{"und-bt", "dz-Tibt-BT"},
    language_tag_expansion{"und-bugi", "bug-Bugi-ID"},
    language_tag_expansion{"und-buhd", "bku-Buhd-PH"},
    language_tag_expansion{"und-bv", "und-Latn-BV"},
    language_tag_expansion{"und-by", "be-Cyrl-BY"},
    language_tag_expansion{"und-cakm", "ccp-Cakm-BD"},
    language_tag_expansion{"und-cans", "cr-Cans-CA"},
    language_tag_expansion{"und-cari", "xcr-Cari-TR"},
    language_tag_expansion{"und-cd", "sw-Latn-CD"},
    language_tag_expansion{"und-cf", "fr-Latn-CF"},
    language_tag_expansion{"und-cg", "fr-Latn-CG"},
    language_tag_expansion{"und-ch", "de-Latn-CH"},
    language_tag_expansion{"und-cham", "cjm-Cham-VN"},
    language_tag_expansion{"und-cher", "chr-Cher-US"},
    language_tag_expansion{"und-chrs", "xco-Chrs-UZ"},
    language_tag_expansion{"und-ci", "fr-Latn-CI"},
    language_tag_expansion{"und-cl", "es-Latn-CL"},
    language_tag_expansion{"und-cm", "fr-Latn-CM"},
    language_tag_expansion{"und-cn", "zh-Hans-CN"},
    language_tag_expansion{"und-co", "es-Latn-CO"},
    language_tag_expansion{"und-copt", "cop-Copt-EG"},
    language_tag_expansion{"und-cp", "und-Latn-CP"},
    language_tag_expansion{"und-cpmn", "und-Cpmn-CY"},
    language_tag_expansion{"und-cprt", "grc-Cprt-CY"},
    language_tag_expansion{"und-cr", "es-Latn-CR"},
    language_tag_expansion{"und-cu", "es-Latn-CU"},
    language_tag_expansion{"und-cv", "pt-Latn-CV"},
    language_tag_expansion{"und-cw", "pap-Latn-CW"},
    language_tag_expansion{"und-cy", "el-Grek-CY"},
    language_tag_expansion{"und-cyrl", "ru-Cyrl-RU"},
    language_tag_expansion{"und-cyrl-al", "mk-Cyrl-AL"},
    language_tag_expansion{"und-cyrl-ba", "sr-Cyrl-BA"},
    language_tag_expansion{"und-cyrl-ge", "os-Cyrl-GE"},
    language_tag_expansion{"und-cyrl-gr", "mk-Cyrl-GR"},
    language_tag_expansion{"und-cyrl-md", "uk-Cyrl-MD"},
    language_tag_expansion{"und-cyrl-ro", "bg-Cyrl-RO"},
    language_tag_expansion{"und-cyrl-sk", "uk-Cyrl-SK"},
    language_tag_expansion{"und-cyrl-tr", "kbd-Cyrl-TR"},
    language_tag_expansion{"und-cyrl-xk", "sr-Cyrl-XK"},
    language_tag_expansion{"und-cz", "cs-Latn-CZ"},
    language_tag_expansion{"und-de", "de-Latn-DE"},
    language_tag_expansion{"und-deva", "hi-Deva-IN"},
    language_tag_expansion{"und-deva-bt", "ne-Deva-BT"},
    language_tag_expansion{"und-deva-fj", "hif-Deva-FJ"},
    language_tag_expansion{"und-deva-mu", "bho-Deva-MU"},
    language_tag_expansion{"und-deva-pk", "btv-Deva-PK"},
    language_tag_expansion{"und-diak", "dv-Diak-MV"},
    language_tag_expansion{"und-dj", "aa-Latn-DJ"},
    language_tag_expansion{"und-dk", "da-Latn-DK"},
    language_tag_expansion{"und-do", "es-Latn-DO"},
    language_tag_expansion{"und-dogr", "doi-Dogr-IN"},
    language_tag_expansion{"und-dupl", "fr-Dupl-FR"},
    language_tag_expansion{"und-dz", "ar-Arab-DZ"},
    language_tag_expansion{"und-ea", "es-Latn-EA"},
    language_tag_expansion{"und-ec", "es-Latn-EC"},
    language_tag_expansion{"und-ee", "et-Latn-EE"},
    language_tag_expansion{"und-eg", "ar-Arab-EG"},
    language_tag_expansion{"und-egyp", "egy-Egyp-EG"},
    language_tag_expansion{"und-eh", "ar-Arab-EH"},
    language_tag_expansion{"und-elba", "sq-Elba-AL"},
    language_tag_expansion{"und-elym", "arc-Elym-IR"},
    language_tag_expansion{"und-er", "ti-Ethi-ER"},
    language_tag_expansion{"und-es", "es-Latn-ES"},
    language_tag_expansion{"und-et", "am-Ethi-ET"},
    language_tag_expansion{"und-ethi", "am-Ethi-ET"},
    language_tag_expansion{"und-eu", "en-Latn-IE"},
    language_tag_expansion{"und-ez", "de-Latn-EZ"},
    language_tag_expansion{"und-fi", "fi-Latn-FI"},
    language_tag_expansion{"und-fo", "fo-Latn-FO"},
    language_tag_expansion{"und-fr", "fr-Latn-FR"},
    language_tag_expansion{"und-ga", "fr-Latn-GA"},
    language_tag_expansion{"und-ge", "ka-Geor-GE"},
    language_tag_expansion{"und-geor", "ka-Geor-GE"},
    language_tag_expansion{"und-gf", "fr-Latn-GF"},
    language_tag_expansion{"und-gh", "ak-Latn-GH"},
    language_tag_expansion{"und-gl", "kl-Latn-GL"},
    language_tag_expansion{"und-glag", "cu-Glag-BG"},
    language_tag_expansion{"und-gn", "fr-Latn-GN"},
    language_tag_expansion{"und-gong", "wsg-Gong-IN"},
    language_tag_expansion{"und-gonm", "esg-Gonm-IN"},
    language_tag_expansion{"und-goth", "got-Goth-UA"},
    language_tag_expansion{"und-gp", "fr-Latn-GP"},
    language_tag_expansion{"und-gq", "es-Latn-GQ"},
    language_tag_expansion{"und-gr", "el-Grek-GR"},
    language_tag_expansion{"und-gran", "sa-Gran-IN"},
    language_tag_expansion{"und-grek", "el-Grek-GR"},
    language_tag_expansion{"und-grek-tr", "bgx-Grek-TR"},
    language_tag_expansion{"und-gs", "und-Latn-GS"},
    language_tag_expansion{"und-gt", "es-Latn-GT"},
    language_tag_expansion{"und-gujr", "gu-Gujr-IN"},
    language_tag_expansion{"und-guru", "pa-Guru-IN"},
    language_tag_expansion{"und-gw", "pt-Latn-GW"},
    language_tag_expansion{"und-hanb", "zh-Hanb-TW"},
    language_tag_expansion{"und-hang", "ko-Hang-KR"},
    language_tag_expansion{"und-hani", "zh-Hani-CN"},
    language_tag_expansion{"und-hano", "hnn-Hano-PH"},
    language_tag_expansion{"und-hans", "zh-Hans-CN"},
    language_tag_expansion{"und-hant", "zh-Hant-TW"},
    language_tag_expansion{"und-hebr", "he-Hebr-IL"},
    language_tag_expansion{"und-hebr-ca", "yi-Hebr-CA"},
    language_tag_expansion{"und-hebr-gb", "yi-Hebr-GB"},
    language_tag_expansion{"und-hebr-se", "yi-Hebr-SE"},
    language_tag_expansion{"und-hebr-ua", "yi-Hebr-UA"},
    language_tag_expansion{"und-hebr-us", "yi-Hebr-US"},
    language_tag_expansion{"und-hira", "ja-Hira-JP"},
    language_tag_expansion{"und-hk", "zh-Hant-HK"},
    language_tag_expansion{"und-hluw", "hlu-Hluw-TR"},
    language_tag_expansion{"und-hm", "und-Latn-HM"},
    language_tag_expansion{"und-hmng", "hnj-Hmng-LA"},
    language_tag_expansion{"und-hmnp", "hnj-Hmnp-US"},
    language_tag_expansion{"und-hn", "es-Latn-HN"},
    language_tag_expansion{"und-hr", "hr-Latn-HR"},
    language_tag_expansion{"und-ht", "ht-Latn-HT"},
    language_tag_expansion{"und-hu", "hu-Latn-HU"},
    language_tag_expansion{"und-hung", "hu-Hung-HU"},
    language_tag_expansion{"und-ic", "es-Latn-IC"},
    language_tag_expansion{"und-id", "id-Latn-ID"},
    language_tag_expansion{"und-il", "he-Hebr-IL"},
    language_tag_expansion{"und-in", "hi-Deva-IN"},
    language_tag_expansion{"und-iq", "ar-Arab-IQ"},
    language_tag_expansion{"und-ir", "fa-Arab-IR"},
    language_tag_expansion{"und-is", "is-Latn-IS"},
    language_tag_expansion{"und-it", "it-Latn-IT"},
    language_tag_expansion{"und-ital", "ett-Ital-IT"},
    language_tag_expansion{"und-jamo", "ko-Jamo-KR"},
    language_tag_expansion{"und-java", "jv-Java-ID"},
    language_tag_expansion{"und-jo", "ar-Arab-JO"},
    language_tag_expansion{"und-jp", "ja-Jpan-JP"},
    language_tag_expansion{"und-jpan", "ja-Jpan-JP"},
    language_tag_expansion{"und-kali", "eky-Kali-MM"},
    language_tag_expansion{"und-kana", "ja-Kana-JP"},
    language_tag_expansion{"und-ke", "sw-Latn-KE"},
    language_tag_expansion{"und-kg", "ky-Cyrl-KG"},
    language_tag_expansion{"und-kh", "km-Khmr-KH"},
    language_tag_expansion{"und-khar", "pra-Khar-PK"},
    language_tag_expansion{"und-khmr", "km-Khmr-KH"},
    language_tag_expansion{"und-khoj", "sd-Khoj-IN"},
    language_tag_expansion{"und-kits", "zkt-Kits-CN"},
    language_tag_expansion{"und-km", "ar-Arab-KM"},
    language_tag_expansion{"und-knda", "kn-Knda-IN"},
    language_tag_expansion{"und-kore", "ko-Kore-KR"},
    language_tag_expansion{"und-kp", "ko-Kore-KP"},
    language_tag_expansion{"und-kr", "ko-Kore-KR"},
    language_tag_expansion{"und-kthi", "bho-Kthi-IN"},
    language_tag_expansion{"und-kw", "ar-Arab-KW"},
    language_tag_expansion{"und-kz", "ru-Cyrl-KZ"},
    language_tag_expansion{"und-la", "lo-Laoo-LA"},
    language_tag_expansion{"und-lana", "nod-Lana-TH"},
    language_tag_expansion{"und-laoo", "lo-Laoo-LA"},
    language_tag_expansion{"und-latn-af", "tk-Latn-AF"},
    language_tag_expansion{"und-latn-am", "ku-Latn-AM"},
    language_tag_expansion{"und-latn-cn", "za-Latn-CN"},
    language_tag_expansion{"und-latn-cy", "tr-Latn-CY"},
    language_tag_expansion{"und-latn-dz", "fr-Latn-DZ"},
    language_tag_expansion{"und-latn-et", "en-Latn-ET"},
    language_tag_expansion{"und-latn-ge", "ku-Latn-GE"},
    language_tag_expansion{"und-latn-ir", "tk-Latn-IR"},
    language_tag_expansion{"und-latn-km", "fr-Latn-KM"},
    language_tag_expansion{"und-latn-ma", "fr-Latn-MA"},
    language_tag_expansion{"und-latn-mk", "sq-Latn-MK"},
    language_tag_expansion{"und-latn-mm", "kac-Latn-MM"},
    language_tag_expansion{"und-latn-mo", "pt-Latn-MO"},
    language_tag_expansion{"und-latn-mr", "fr-Latn-MR"},
    language_tag_expansion{"und-latn-ru", "krl-Latn-RU"},
    language_tag_expansion{"und-latn-sy", "fr-Latn-SY"},
    language_tag_expansion{"und-latn-tn", "fr-Latn-TN"},
    language_tag_expansion{"und-latn-tw", "trv-Latn-TW"},
    language_tag_expansion{"und-latn-ua", "pl-Latn-UA"},
    language_tag_expansion{"und-lb", "ar-Arab-LB"},
    language_tag_expansion{"und-lepc", "lep-Lepc-IN"},
    language_tag_expansion{"und-li", "de-Latn-LI"},
    language_tag_expansion{"und-limb", "lif-Limb-IN"},
    language_tag_expansion{"und-lina", "lab-Lina-GR"},
    language_tag_expansion{"und-linb", "grc-Linb-GR"},
    language_tag_expansion{"und-lisu", "lis-Lisu-CN"},
    language_tag_expansion{"und-lk", "si-Sinh-LK"},
    language_tag_expansion{"und-ls", "st-Latn-LS"},
    language_tag_expansion{"und-lt", "lt-Latn-LT"},
    language_tag_expansion{"und-lu", "fr-Latn-LU"},
    language_tag_expansion{"und-lv", "lv-Latn-LV"},
    language_tag_expansion{"und-ly", "ar-Arab-LY"},
    language_tag_expansion{"und-lyci", "xlc-Lyci-TR"},
    language_tag_expansion{"und-lydi", "xld-Lydi-TR"},
    language_tag_expansion{"und-ma", "ar-Arab-MA"},
    language_tag_expansion{"und-mahj", "hi-Mahj-IN"},
    language_tag_expansion{"und-maka", "mak-Maka-ID"},
    language_tag_expansion{"und-mand", "myz-Mand-IR"},
    language_tag_expansion{"und-mani", "xmn-Mani-CN"},
    language_tag_expansion{"und-marc", "bo-Marc-CN"},
    language_tag_expansion{"und-mc", "fr-Latn-MC"},
    language_tag_expansion{"und-md", "ro-Latn-MD"},
    language_tag_expansion{"und-me", "sr-Latn-ME"},
    language_tag_expansion{"und-medf", "dmf-Medf-NG"},
    language_tag_expansion{"und-mend", "men-Mend-SL"},
    language_tag_expansion{"und-merc", "xmr-Merc-SD"},
    language_tag_expansion{"und-mero", "xmr-Mero-SD"},
    language_tag_expansion{"und-mf", "fr-Latn-MF"},
    language_tag_expansion{"und-mg", "mg-Latn-MG"},
    language_tag_expansion{"und-mk", "mk-Cyrl-MK"},
    language_tag_expansion{"und-ml", "bm-Latn-ML"},
    language_tag_expansion{"und-mlym", "ml-Mlym-IN"},
    language_tag_expansion{"und-mm", "my-Mymr-MM"},
    language_tag_expansion{"und-mn", "mn-Cyrl-MN"},
    language_tag_expansion{"und-mo", "zh-Hant-MO"},
    language_tag_expansion{"und-modi", "mr-Modi-IN"},
    language_tag_expansion{"und-mong", "mn-Mong-CN"},
    language_tag_expansion{"und-mq", "fr-Latn-MQ"},
    language_tag_expansion{"und-mr", "ar-Arab-MR"},
    language_tag_expansion{"und-mroo", "mro-Mroo-BD"},
    language_tag_expansion{"und-mt", "mt-Latn-MT"},
    language_tag_expansion{"und-mtei", "mni-Mtei-IN"},
    language_tag_expansion{"und-mu", "mfe-Latn-MU"},
    language_tag_expansion{"und-mult", "skr-Mult-PK"},
    language_tag_expansion{"und-mv", "dv-Thaa-MV"},
    language_tag_expansion{"und-mx", "es-Latn-MX"},
    language_tag_expansion{"und-my", "ms-Latn-MY"},
    language_tag_expansion{"und-mymr", "my-Mymr-MM"},
    language_tag_expansion{"und-mymr-in", "kht-Mymr-IN"},
    language_tag_expansion{"und-mymr-th", "mnw-Mymr-TH"},
    language_tag_expansion{"und-mz", "pt-Latn-MZ"},
    language_tag_expansion{"und-na", "af-Latn-NA"},
    language_tag_expansion{"und-nand", "sa-Nand-IN"},
    language_tag_expansion{"und-narb", "xna-Narb-SA"},
    language_tag_expansion{"und-nbat", "arc-Nbat-JO"},
    language_tag_expansion{"und-nc", "fr-Latn-NC"},
    language_tag_expansion{"und-ne", "ha-Latn-NE"},
    language_tag_expansion{"und-newa", "new-Newa-NP"},
    language_tag_expansion{"und-ni", "es-Latn-NI"},
    language_tag_expansion{"und-nkoo", "man-Nkoo-GN"},
    language_tag_expansion{"und-nl", "nl-Latn-NL"},
    language_tag_expansion{"und-no", "nb-Latn-NO"},
    language_tag_expansion{"und-np", "ne-Deva-NP"},
    language_tag_expansion{"und-nshu", "zhx-Nshu-CN"},
    language_tag_expansion{"und-ogam", "sga-Ogam-IE"},
    language_tag_expansion{"und-olck", "sat-Olck-IN"},
    language_tag_expansion{"und-om", "ar-Arab-OM"},
    language_tag_expansion{"und-orkh", "otk-Orkh-MN"},
    language_tag_expansion{"und-orya", "or-Orya-IN"},
    language_tag_expansion{"und-osge", "osa-Osge-US"},
    language_tag_expansion{"und-osma", "so-Osma-SO"},
    language_tag_expansion{"und-ougr", "oui-Ougr-143"},
    language_tag_expansion{"und-pa", "es-Latn-PA"},
    language_tag_expansion{"und-palm", "arc-Palm-SY"},
    language_tag_expansion{"und-pauc", "ctd-Pauc-MM"},
    language_tag_expansion{"und-pe", "es-Latn-PE"},
    language_tag_expansion{"und-perm", "kv-Perm-RU"},
    language_tag_expansion{"und-pf", "fr-Latn-PF"},
    language_tag_expansion{"und-pg", "tpi-Latn-PG"},
    language_tag_expansion{"und-ph", "fil-Latn-PH"},
    language_tag_expansion{"und-phag", "lzh-Phag-CN"},
    language_tag_expansion{"und-phli", "pal-Phli-IR"},
    language_tag_expansion{"und-phlp", "pal-Phlp-CN"},
    language_tag_expansion{"und-phnx", "phn-Phnx-LB"},
    language_tag_expansion{"und-pk", "ur-Arab-PK"},
    language_tag_expansion{"und-pl", "pl-Latn-PL"},
    language_tag_expansion{"und-plrd", "hmd-Plrd-CN"},
    language_tag_expansion{"und-pm", "fr-Latn-PM"},
    language_tag_expansion{"und-pr", "es-Latn-PR"},
    language_tag_expansion{"und-prti", "xpr-Prti-IR"},
    language_tag_expansion{"und-ps", "ar-Arab-PS"},
    language_tag_expansion{"und-pt", "pt-Latn-PT"},
    language_tag_expansion{"und-pw", "pau-Latn-PW"},
    language_tag_expansion{"und-py", "gn-Latn-PY"},
    language_tag_expansion{"und-qa", "ar-Arab-QA"},
    language_tag_expansion{"und-qo", "en-Latn-DG"},
    language_tag_expansion{"und-re", "fr-Latn-RE"},
    language_tag_expansion{"und-rjng", "rej-Rjng-ID"},
    language_tag_expansion{"und-ro", "ro-Latn-RO"},
    language_tag_expansion{"und-rohg", "rhg-Rohg-MM"},
    language_tag_expansion{"und-rs", "sr-Cyrl-RS"},
    language_tag_expansion{"und-ru", "ru-Cyrl-RU"},
    language_tag_expansion{"und-runr", "non-Runr-SE"},
    language_tag_expansion{"und-rw", "rw-Latn-RW"},
    language_tag_expansion{"und-sa", "ar-Arab-SA"},
    language_tag_expansion{"und-samr", "smp-Samr-IL"},
    language_tag_expansion{"und-sarb", "xsa-Sarb-YE"},
    language_tag_expansion{"und-saur", "saz-Saur-IN"},
    language_tag_expansion{"und-sc", "fr-Latn-SC"},
    language_tag_expansion{"und-sd", "ar-Arab-SD"},
    language_tag_expansion{"und-se", "sv-Latn-SE"},
    language_tag_expansion{"und-sgnw", "ase-Sgnw-US"},
    language_tag_expansion{"und-shaw", "en-Shaw-GB"},
    language_tag_expansion{"und-shrd", "sa-Shrd-IN"},
    language_tag_expansion{"und-si", "sl-Latn-SI"},
    language_tag_expansion{"und-sidd", "sa-Sidd-IN"},
    language_tag_expansion{"und-sind", "sd-Sind-IN"},
    language_tag_expansion{"und-sinh", "si-Sinh-LK"},
    language_tag_expansion{"und-sj", "nb-Latn-SJ"},
    language_tag_expansion{"und-sk", "sk-Latn-SK"},
    language_tag_expansion{"und-sm", "it-Latn-SM"},
    language_tag_expansion{"und-sn", "fr-Latn-SN"},
    language_tag_expansion{"und-so", "so-Latn-SO"},
    language_tag_expansion{"und-sogd", "sog-Sogd-UZ"},
    language_tag_expansion{"und-sogo", "sog-Sogo-UZ"},
    language_tag_expansion{"und-sora", "srb-Sora-IN"},
    language_tag_expansion{"und-soyo", "cmg-Soyo-MN"},
    language_tag_expansion{"und-sr", "nl-Latn-SR"},
    language_tag_expansion{"und-st", "pt-Latn-ST"},
    language_tag_expansion{"und-sund", "su-Sund-ID"},
    language_tag_expansion{"und-sv", "es-Latn-SV"},
    language_tag_expansion{"und-sy", "ar-Arab-SY"},
    language_tag_expansion{"und-sylo", "syl-Sylo-BD"},
    language_tag_expansion{"und-syrc", "syr-Syrc-IQ"},
    language_tag_expansion{"und-tagb", "tbw-Tagb-PH"},
    language_tag_expansion{"und-takr", "doi-Takr-IN"},
    language_tag_expansion{"und-tale", "tdd-Tale-CN"},
    language_tag_expansion{"und-talu", "khb-Talu-CN"},
    language_tag_expansion{"und-taml", "ta-Taml-IN"},
    language_tag_expansion{"und-tang", "txg-Tang-CN"},
    language_tag_expansion{"und-tavt", "blt-Tavt-VN"},
    language_tag_expansion{"und-td", "fr-Latn-TD"},
    language_tag_expansion{"und-telu", "te-Telu-IN"},
    language_tag_expansion{"und-tf", "fr-Latn-TF"},
    language_tag_expansion{"und-tfng", "zgh-Tfng-MA"},
    language_tag_expansion{"und-tg", "fr-Latn-TG"},
    language_tag_expansion{"und-tglg", "fil-Tglg-PH"},
    language_tag_expansion{"und-th", "th-Thai-TH"},
    language_tag_expansion{"und-thaa", "dv-Thaa-MV"},
    language_tag_expansion{"und-thai", "th-Thai-TH"},
    language_tag_expansion{"und-thai-cn", "lcp-Thai-CN"},
    language_tag_expansion{"und-thai-kh", "kdt-Thai-KH"},
    language_tag_expansion{"und-thai-la", "kdt-Thai-LA"},
    language_tag_expansion{"und-tibt", "bo-Tibt-CN"},
    language_tag_expansion{"und-tirh", "mai-Tirh-IN"},
    language_tag_expansion{"und-tj", "tg-Cyrl-TJ"},
    language_tag_expansion{"und-tk", "tkl-Latn-TK"},
    language_tag_expansion{"und-tl", "pt-Latn-TL"},
    language_tag_expansion{"und-tm", "tk-Latn-TM"},
    language_tag_expansion{"und-tn", "ar-Arab-TN"},
    language_tag_expansion{"und-tnsa", "nst-Tnsa-IN"},
    language_tag_expansion{"und-to", "to-Latn-TO"},
    language_tag_expansion{"und-toto", "txo-Toto-IN"},
    language_tag_expansion{"und-tr", "tr-Latn-TR"},
    language_tag_expansion{"und-tv", "tvl-Latn-TV"},
    language_tag_expansion{"und-tw", "zh-Hant-TW"},
    language_tag_expansion{"und-tz", "sw-Latn-TZ"},
    language_tag_expansion{"und-ua", "uk-Cyrl-UA"},
    language_tag_expansion{"und-ug", "sw-Latn-UG"},
    language_tag_expansion{"und-ugar", "uga-Ugar-SY"},
    language_tag_expansion{"und-uy", "es-Latn-UY"},
    language_tag_expansion{"und-uz", "uz-Latn-UZ"},
    language_tag_expansion{"und-va", "it-Latn-VA"},
    language_tag_expansion{"und-vaii", "vai-Vaii-LR"},
    language_tag_expansion{"und-ve", "es-Latn-VE"},
    language_tag_expansion{"und-vith", "sq-Vith-AL"},
    language_tag_expansion{"und-vn", "vi-Latn-VN"},
    language_tag_expansion{"und-vu", "bi-Latn-VU"},
    language_tag_expansion{"und-wara", "hoc-Wara-IN"},
    language_tag_expansion{"und-wcho", "nnp-Wcho-IN"},
    language_tag_expansion{"und-wf", "fr-Latn-WF"},
    language_tag_expansion{"und-ws", "sm-Latn-WS"},
    language_tag_expansion{"und-xk", "sq-Latn-XK"},
    language_tag_expansion{"und-xpeo", "peo-Xpeo-IR"},
    language_tag_expansion{"und-xsux", "akk-Xsux-IQ"},
    language_tag_expansion{"und-ye", "ar-Arab-YE"},
    language_tag_expansion{"und-yezi", "ku-Yezi-GE"},
    language_tag_expansion{"und-yiii", "ii-Yiii-CN"},
    language_tag_expansion{"und-yt", "fr-Latn-YT"},
    language_tag_expansion{"und-zanb", "cmg-Zanb-MN"},
    language_tag_expansion{"und-zw", "sn-Latn-ZW"},
    language_tag_expansion{"unr", "unr-Beng-IN"},
    language_tag_expansion{"unr-deva", "unr-Deva-NP"},
    language_tag_expansion{"unr-np", "unr-Deva-NP"},
    language_tag_expansion{"unx", "unx-Beng-IN"},
    language_tag_expansion{"uok", "uok-Latn-ZZ"},
    language_tag_expansion{"ur", "ur-Arab-PK"},
    language_tag_expansion{"uri", "uri-Latn-ZZ"},
    language_tag_expansion{"urt", "urt-Latn-ZZ"},
    language_tag_expansion{"urw", "urw-Latn-ZZ"},
    language_tag_expansion{"usa", "usa-Latn-ZZ"},
    language_tag_expansion{"uth", "uth-Latn-ZZ"},
    language_tag_expansion{"utr", "utr-Latn-ZZ"},
    language_tag_expansion{"uvh", "uvh-Latn-ZZ"},
    language_tag_expansion{"uvl", "uvl-Latn-ZZ"},
    language_tag_expansion{"uz", "uz-Latn-UZ"},
    language_tag_expansion{"uz-af", "uz-Arab-AF"},
    language_tag_expansion{"uz-arab", "uz-Arab-AF"},
    language_tag_expansion{"uz-cn", "uz-Cyrl-CN"},
    language_tag_expansion{"vag", "vag-Latn-ZZ"},
    language_tag_expansion{"vai", "vai-Vaii-LR"},
    language_tag_expansion{"van", "van-Latn-ZZ"},
    language_tag_expansion{"ve", "ve-Latn-ZA"},
    language_tag_expansion{"vec", "vec-Latn-IT"},
    language_tag_expansion{"vep", "vep-Latn-RU"},
    language_tag_expansion{"vi", "vi-Latn-VN"},
    language_tag_expansion{"vic", "vic-Latn-SX"},
    language_tag_expansion{"viv", "viv-Latn-ZZ"},
    language_tag_expansion{"vls", "vls-Latn-BE"},
    language_tag_expansion{"vmf", "vmf-Latn-DE"},
    language_tag_expansion{"vmw", "vmw-Latn-MZ"},
    language_tag_expansion{"vo", "vo-Latn-001"},
    language_tag_expansion{"vot", "vot-Latn-RU"},
    language_tag_expansion{"vro", "vro-Latn-EE"},
    language_tag_expansion{"vun", "vun-Latn-TZ"},
    language_tag_expansion{"vut", "vut-Latn-ZZ"},
    language_tag_expansion{"wa", "wa-Latn-BE"},
    language_tag_expansion{"wae", "wae-Latn-CH"},
    language_tag_expansion{"waj", "waj-Latn-ZZ"},
    language_tag_expansion{"wal", "wal-Ethi-ET"},
    language_tag_expansion{"wan", "wan-Latn-ZZ"},
    language_tag_expansion{"war", "war-Latn-PH"},
    language_tag_expansion{"wbp", "wbp-Latn-AU"},
    language_tag_expansion{"wbq", "wbq-Telu-IN"},
    language_tag_expansion{"wbr", "wbr-Deva-IN"},
    language_tag_expansion{"wci", "wci-Latn-ZZ"},
    language_tag_expansion{"wer", "wer-Latn-ZZ"},
    language_tag_expansion{"wgi", "wgi-Latn-ZZ"},
    language_tag_expansion{"whg", "whg-Latn-ZZ"},
    language_tag_expansion{"wib", "wib-Latn-ZZ"},
    language_tag_expansion{"wiu", "wiu-Latn-ZZ"},
    language_tag_expansion{"wiv", "wiv-Latn-ZZ"},
    language_tag_expansion{"wja", "wja-Latn-ZZ"},
    language_tag_expansion{"wji", "wji-Latn-ZZ"},
    language_tag_expansion{"wls", "wls-Latn-WF"},
    language_tag_expansion{"wmo", "wmo-Latn-ZZ"},
    language_tag_expansion{"wnc", "wnc-Latn-ZZ"},
    language_tag_expansion{"wni", "wni-Arab-KM"},
    language_tag_expansion{"wnu", "wnu-Latn-ZZ"},
    language_tag_expansion{"wo", "wo-Latn-SN"},
    language_tag_expansion{"wob", "wob-Latn-ZZ"},
    language_tag_expansion{"wos", "wos-Latn-ZZ"},
    language_tag_expansion{"wrs", "wrs-Latn-ZZ"},
    language_tag_expansion{"wsg", "wsg-Gong-IN"},
    language_tag_expansion{"wsk", "wsk-Latn-ZZ"},
    language_tag_expansion{"wtm", "wtm-Deva-IN"},
    language_tag_expansion{"wuu", "wuu-Hans-CN"},
    language_tag_expansion{"wuv", "wuv-Latn-ZZ"},
    language_tag_expansion{"wwa", "wwa-Latn-ZZ"},
    language_tag_expansion{"xav", "xav-Latn-BR"},
    language_tag_expansion{"xbi", "xbi-Latn-ZZ"},
    language_tag_expansion{"xco", "xco-Chrs-UZ"},
    language_tag_expansion{"xcr", "xcr-Cari-TR"},
    language_tag_expansion{"xes", "xes-Latn-ZZ"},
    language_tag_expansion{"xh", "xh-Latn-ZA"},
    language_tag_expansion{"xla", "xla-Latn-ZZ"},
    language_tag_expansion{"xlc", "xlc-Lyci-TR"},
    language_tag_expansion{"xld", "xld-Lydi-TR"},
    language_tag_expansion{"xmf", "xmf-Geor-GE"},
    language_tag_expansion{"xmn", "xmn-Mani-CN"},
    language_tag_expansion{"xmr", "xmr-Merc-SD"},
    language_tag_expansion{"xna", "xna-Narb-SA"},
    language_tag_expansion{"xnr", "xnr-Deva-IN"},
    language_tag_expansion{"xog", "xog-Latn-UG"},
    language_tag_expansion{"xon", "xon-Latn-ZZ"},
    language_tag_expansion{"xpr", "xpr-Prti-IR"},
    language_tag_expansion{"xrb", "xrb-Latn-ZZ"},
    language_tag_expansion{"xsa", "xsa-Sarb-YE"},
    language_tag_expansion{"xsi", "xsi-Latn-ZZ"},
    language_tag_expansion{"xsm", "xsm-Latn-ZZ"},
    language_tag_expansion{"xsr", "xsr-Deva-NP"},
    language_tag_expansion{"xwe", "xwe-Latn-ZZ"},
    language_tag_expansion{"yam", "yam-Latn-ZZ"},
    language_tag_expansion{"yao", "yao-Latn-MZ"},
    language_tag_expansion{"yap", "yap-Latn-FM"},
    language_tag_expansion{"yas", "yas-Latn-ZZ"},
    language_tag_expansion{"yat", "yat-Latn-ZZ"},
    language_tag_expansion{"yav", "yav-Latn-CM"},
    language_tag_expansion{"yay", "yay-Latn-ZZ"},
    language_tag_expansion{"yaz", "yaz-Latn-ZZ"},
    language_tag_expansion{"yba", "yba-Latn-ZZ"},
    language_tag_expansion{"ybb", "ybb-Latn-CM"},
    language_tag_expansion{"yby", "yby-Latn-ZZ"},
    language_tag_expansion{"yer", "yer-Latn-ZZ"},
    language_tag_expansion{"ygr", "ygr-Latn-ZZ"},
    language_tag_expansion{"ygw", "ygw-Latn-ZZ"},
    language_tag_expansion{"yi", "yi-Hebr-001"},
    language_tag_expansion{"yko", "yko-Latn-ZZ"},
    language_tag_expansion{"yle", "yle-Latn-ZZ"},
    language_tag_expansion{"ylg", "ylg-Latn-ZZ"},
    language_tag_expansion{"yll", "yll-Latn-ZZ"},
    language_tag_expansion{"yml", "yml-Latn-ZZ"},
    language_tag_expansion{"yo", "yo-Latn-NG"},
    language_tag_expansion{"yon", "yon-Latn-ZZ"},
    language_tag_expansion{"yrb", "yrb-Latn-ZZ"},
    language_tag_expansion{"yre", "yre-Latn-ZZ"},
    language_tag_expansion{"yrl", "yrl-Latn-BR"},
    language_tag_expansion{"yss", "yss-Latn-ZZ"},
    language_tag_expansion{"yua", "yua-Latn-MX"},
    language_tag_expansion{"yue", "yue-Hant-HK"},
    language_tag_expansion{"yue-cn", "yue-Hans-CN"},
    language_tag_expansion{"yue-hans", "yue-Hans-CN"},
    language_tag_expansion{"yuj", "yuj-Latn-ZZ"},
    language_tag_expansion{"yut", "yut-Latn-ZZ"},
    language_tag_expansion{"yuw", "yuw-Latn-ZZ"},
    language_tag_expansion{"za", "za-Latn-CN"},
    language_tag_expansion{"zag", "zag-Latn-SD"},
    language_tag_expansion{"zdj", "zdj-Arab-KM"},
    language_tag_expansion{"zea", "zea-Latn-NL"},
    language_tag_expansion{"zgh", "zgh-Tfng-MA"},
    language_tag_expansion{"zh", "zh-Hans-CN"},
    language_tag_expansion{"zh-au", "zh-Hant-AU"},
    language_tag_expansion{"zh-bn", "zh-Hant-BN"},
    language_tag_expansion{"zh-bopo", "zh-Bopo-TW"},
    language_tag_expansion{"zh-gb", "zh-Hant-GB"},
    language_tag_expansion{"zh-gf", "zh-Hant-GF"},
    language_tag_expansion{"zh-guoyu", "cmn"}, // grandfathered
    language_tag_expansion{"zh-hakka", "hak-Hans-CN"}, // grandfathered
    language_tag_expansion{"zh-hanb", "zh-Hanb-TW"},
    language_tag_expansion{"zh-hant", "zh-Hant-TW"},
    language_tag_expansion{"zh-hk", "zh-Hant-HK"},
    language_tag_expansion{"zh-id", "zh-Hant-ID"},
    language_tag_expansion{"zh-min-nan", "nan-Hans-CN"}, // grandfathered
    language_tag_expansion{"zh-mo", "zh-Hant-MO"},
    language_tag_expansion{"zh-pa", "zh-Hant-PA"},
    language_tag_expansion{"zh-pf", "zh-Hant-PF"},
    language_tag_expansion{"zh-ph", "zh-Hant-PH"},
    language_tag_expansion{"zh-sr", "zh-Hant-SR"},
    language_tag_expansion{"zh-th", "zh-Hant-TH"},
    language_tag_expansion{"zh-tw", "zh-Hant-TW"},
    language_tag_expansion{"zh-us", "zh-Hant-US"},
    language_tag_expansion{"zh-vn", "zh-Hant-VN"},
    language_tag_expansion{"zhx", "zhx-Nshu-CN"},
    language_tag_expansion{"zh-xiang", "hsn-Hans-CN"}, // grandfathered
    language_tag_expansion{"zia", "zia-Latn-ZZ"},
    language_tag_expansion{"zkt", "zkt-Kits-CN"},
    language_tag_expansion{"zlm", "zlm-Latn-TG"},
    language_tag_expansion{"zmi", "zmi-Latn-MY"},
    language_tag_expansion{"zne", "zne-Latn-ZZ"},
    language_tag_expansion{"zu", "zu-Latn-ZA"},
#endif
    language_tag_expansion{"zza", "zza-Latn-TR"}};
// clang-format on

static std::optional<std::string_view> expand_language_tag(std::string_view from) noexcept
{
    hilet from_lower = to_lower(from);
    hilet it = std::lower_bound(
        language_tag_expansions.begin(),
        language_tag_expansions.end(),
        std::string_view{from_lower},
        [](hilet& item, hilet& value) {
            return item.from < value;
        });

    if (it != language_tag_expansions.end() and it->from == from) {
        return it->to;
    } else {
        return {};
    }
}

[[nodiscard]] language_tag language_tag::parse(std::string_view str)
{
    auto language = iso_639{};
    auto script = iso_15924{};
    auto region = iso_3166{};

    char extension_first_char = 0;

    // Replace underscores to dashes, since invalid language-tags do exist in the real world.
    auto str_ = std::string{str};
    for (auto& c : str_) {
        if (c == '_') {
            c = '-';
        }
    }
    str = std::string_view{str_};

    for (hilet element : std::views::split(str, std::string_view{"-"})) {
        if (extension_first_char) {
            // Once inside the extensions portion of a language tag you can no
            // longer determine validity based on just the element size.
            ;

        } else if (not language) {
            hi_parse_check(
                (element.size() == 2 or element.size() == 3) and is_alpha(element),
                "First element of a language tag must be a ISO-639 2 or 3 letter language code, got '{}'",
                str);
            // 2 or 3 letter non-optional ISO-639 language code.
            language = {element};

        } else {
            if (not script and not region and element.size() == 3 and is_alpha(element)) {
                // Up to 3 optional 3 letter extended language codes.
                // Ignore these for backward compatibility.
                ;

            } else if (not script and not region and element.size() == 4 and is_alpha(element)) {
                // The language code may be followed by a 4 letter script code.
                script = {element};

            } else if (not region and (element.size() == 2 and is_alpha(element))) {
                // The language code or script code may also be followed by a 2 letter country code.
                region = {element};

            } else if (not region and ((element.size() == 2 or element.size() == 3) and is_digit(element))) {
                // The language code or script code may also be followed by a 2 or 3 digit country code.
                region = {element};

            } else if ((element.size() >= 5 and element.size() <= 8) or (element.size() == 4 and is_digit(element.front()))) {
                // A variant has 5 to 8 letters or a 4 digit + letters code.
                ;

            } else if (element.size() == 1) {
                // Start of an extension. We do not differentiate with private-use indicator.
                extension_first_char = element.front();

            } else {
                throw parse_error(
                    std::format("Unexpected element '{}' while parsing language tag '{}'", std::string_view{element}, str));
            }
        }
    }

    return language_tag{language, script, region};
}

[[nodiscard]] language_tag language_tag::expand() const noexcept
{
    auto r = *this;

    if (script and region) {
        return r;
    }

    if (auto from_language = expand_language_tag(r.language.code())) {
        auto from_language_tag = parse(*from_language);

        if (not r.script and from_language_tag.script) {
            r.script = from_language_tag.script;
        }
        if (not r.region and from_language_tag.region) {
            r.region = from_language_tag.region;
        }
    }

    if (script and region) {
        return r;
    }

    if (auto from_region = expand_language_tag(std::string{"und-"} + std::string{r.region.code2()})) {
        auto from_region_tag = parse(*from_region);

        if (not r.script and from_region_tag.script) {
            r.script = from_region_tag.script;
        }
    }

    return r;
}

language_tag::language_tag(std::string_view str) : language(), script(), region()
{
    // First do an initial pass over the expansion table to convert likely languages.
    // For example "nl" -> "nl-Latn-NL", "nl-BE" -> "nl-BE".
    if (auto expanded_str = expand_language_tag(str)) {
        str = *expanded_str;
    }

    *this = parse(str).expand();
}

[[nodiscard]] std::vector<language_tag> variants(std::vector<language_tag> languages)
{
    auto tmp = std::vector<std::vector<language_tag>>{};

    for (hilet& language : languages) {
        auto& lang_tmp = tmp.emplace_back();
        for (hilet& variant : language.all_variants()) {
            lang_tmp.push_back(variant);
        }
    }

    for (auto it = tmp.rbegin(); it != tmp.rend(); ++it) {
        // Remove duplicates in previous language-variant lists.
        for (auto jt = it + 1; jt != tmp.rend(); ++jt) {
            for (hilet& tag : *it) {
                std::erase(*jt, tag);
            }
        }
    }

    auto r = std::vector<language_tag>{};

    hilet count = std::accumulate(tmp.begin(), tmp.end(), 0_uz, [](hilet& value, hilet& item) {
        return value + item.size();
    });
    r.reserve(count);

    for (hilet& variants : tmp) {
        for (hilet& tag : variants) {
            r.push_back(tag);
        }
    }

    return r;
}

} // namespace hi::inline v1